# -*- coding: utf-8 -*-
# =============================================================================
#
# Created on Tue Jan  7 12:14:11 2025
#
# @author: kenny
# =============================================================================
"""
Macros for convenience when parsing qasm code.
"""


# =============================================================================
# #copy and paste into AstInclude and appropriate places in ast2dqc_circuit_tests
# # then DELETE standard lib when finished making
# #macros for gates not in it
# standard_lib = {"u3" : gates.INSTR_U, #alias of U native gate
#                  "u2" : lambda phi, lambda_var : gates.INSTR_U(np.pi/2, phi, lambda_var),
#                  "u1" : lambda lambda_var : gates.INSTR_U(0, 0, lambda_var),
#                  "cx" : instr.INSTR_CNOT,
#                  "id" : gates.INSTR_IDENTITY,
#                  "u0" : gates.INSTR_U(0, 0, 0),
#                  "u" : gates.INSTR_U, #alias of U and u3
#                  "p" : lambda lambda_var : gates.INSTR_U(0, 0, lambda_var), #alias of u1
#                  "x" : instr.INSTR_X,
#                  "y" : instr.INSTR_Y,
#                  "z" : instr.INSTR_Z,
#                  "h" : instr.INSTR_H,
#                  "s" : instr.INSTR_S,
#                  "sdg" : gates.INSTR_S_DAGGER,
#                  "t" : instr.INSTR_T,
#                  "tdg" : gates.INSTR_T_DAGGER,
#                  "rx" : lambda theta : gates.INSTR_U(theta, -np.pi/2, np.pi/2),
#                  "ry" : lambda theta : gates.INSTR_U(theta, 0, 0),
#                  "rz" : gates.instrNop_RZ,
#                  "sx" : gates.instrNop_SX,
#                  "sxdg" : gates.instrNop_SXDG,
#                  "cz" : instr.INSTR_CZ,
#                  "cy" : gates.INSTR_CY,
#                  "ch" : gates.INSTR_CH,
#                  "crx" : lambda theta : gates.INSTR_U(theta, -np.pi/2, np.pi/2,
#                                        controlled=True),
#                  "cry" : lambda theta : gates.INSTR_U(theta, 0, 0,
#                                                       controlled=True),
#                  "crz" : lambda angle : gates.instrNop_RZ(angle,
#                                                           controlled=True),
#                  "cu1" : lambda lambda_var : gates.INSTR_U(0, 0, lambda_var,
#                                                            controlled=True),
#                  "cp" : lambda lambda_var : gates.INSTR_U(0, 0, lambda_var,
#                                                           controlled=True), #alias of cu1
#                  "cu3" : lambda theta, phi, lambda_var : gates.INSTR_U(
#                                      theta, phi, lambda_var, controlled=True),
#                  "csx" : gates.instrNop_CSX}
# =============================================================================

# defining some useful strings
lpar = "("
rpar = ")"
plus = "+"
over2 = "/2"
minus = ""

# BELOW are for the U, CU universal set

# added u0, sx, sxdg, csx


# =============================================================================
# class StandardLibMacros():
#     """ Macros for the gates in the openQASM 2.0 standard library (using the
#         library defined in MQT bench as of November 2023 rather than the less
#         extensive library defined in the original propasal for openQASM 2.0),
#         which are not in
#         the {U, CU} set, (ie, not 1 or 2-qubit gates). I am using this under the
#         assumption that the gate times and error will not be that different between
#         different single qubit gates and that the difference between a CX and
#         CU gate will be negligible relative to the difference between a single
#         -qubit gate and a two-qubit gate.
#     """
#
# =============================================================================

# want list of dicts with entries 'subgate_name', 'subgate_params', 'subgate_args'

# The following macros use poor variable names like a, b, c, etc. This is
# because they were being manually copied by hand from the openQASM 2.0
# specification and it was easier to avoid mistakes when using the same notation
# they did


def cz_macro(a, b):
    """
    Macro for CZ gate in terms of H and CX.

    For internal use by :mod: `~dqc_simulator.software.ast2dqc_circuit`.
    Further compilation will be needed before the output can be interpretted
    by the simulator itself.

    Parameters
    ----------
    a : int
        Control qubit index
    b : int
        Target qubit index

    Returns
    -------
    subgates : list of dicts
        The gates comprising the macro.
    """
    h_gate = {"name": "h", "params": None, "args": [b]}
    cx_gate = {"name": "cx", "params": None, "args": [a, b]}
    subgates = [h_gate, cx_gate, h_gate]
    return subgates


def cy_macro(a, b):
    """
    Macro for CY gate in terms of CX, S and S^dagger.

    For internal use by :mod: `~dqc_simulator.software.ast2dqc_circuit`.
    Further compilation will be needed before the output can be interpretted
    by the simulator itself.

    Parameters
    ----------
    a : int
        Control qubit index
    b : int
        Target qubit index

    Returns
    -------
    subgates : list of dicts
        The gates comprising the macro.
    """
    subgates = [
        {"name": "sdg", "params": None, "args": [b]},
        {"name": "cx", "params": None, "args": [a, b]},
        {"name": "s", "params": None, "args": [b]},
    ]
    return subgates


def swap_macro(parent_arg1, parent_arg2):
    """
    Macro for swap gate in terms of CX gates.

    Parameters
    ----------
    parent_arg1 : str
        The first of the arguments provided in the call of the SWAP gate.
    parent_arg2 : str
        The second argument provided in the call of the SWAP gate.

    Returns
    -------
    list of dict
        List of the subgates in the macro.

    """
    common_entries = {"name": "cx", "params": None}
    subgate1 = {**common_entries, "args": [parent_arg1, parent_arg2]}
    subgate2 = {**common_entries, "args": [parent_arg2, parent_arg1]}
    return [subgate1, subgate2, subgate1]


def ch_macro(a, b):
    """
    Macro for CH gate.

    For internal use by :mod: `~dqc_simulator.software.ast2dqc_circuit`.
    Further compilation will be needed before the output can be interpretted
    by the simulator itself.

    Parameters
    ----------
    a : int
        Control qubit index
    b : int
        Target qubit index

    Returns
    -------
    subgates : list of dicts
        The gates comprising the macro.
    """
    subgates = [
        {"name": "h", "params": None, "args": [b]},
        {"name": "sdg", "params": None, "args": [b]},
        {"name": "cx", "params": None, "args": [a, b]},
        {"name": "h", "params": None, "args": [b]},
        {"name": "t", "params": None, "args": [b]},
        {"name": "cx", "params": None, "args": [a, b]},
        {"name": "t", "params": None, "args": [b]},
        {"name": "h", "params": None, "args": [b]},
        {"name": "s", "params": None, "args": [b]},
        {"name": "x", "params": None, "args": [b]},
        {"name": "s", "params": None, "args": [a]},
    ]
    return subgates


def ccx_macro(a, b, c):
    """
    macro for ccx gate, as defined in MQT bench's version of the openQASM 2.0
    standard library (qelib1.inc)

    Parameters
    ----------
    a : str
        First uninterpretted argument of parent gate.
    b : str
        Second uninterpretted argument of parent gate.
    c : str
        Third uninterpretted argument of parent gate.

    Returns
    -------
    subgates : list of dicts
        The gates comprising the macro.

    """
    cx_common_entries = {"name": "cx", "params": None}
    subgates = [
        {"name": "h", "params": None, "args": [c]},
        {**cx_common_entries, "args": [b, c]},
        {"name": "tdg", "params": None, "args": [c]},
        {**cx_common_entries, "args": [a, c]},
        {"name": "t", "params": None, "args": [c]},
        {**cx_common_entries, "args": [b, c]},
        {"name": "tdg", "params": None, "args": [c]},
        {**cx_common_entries, "args": [a, c]},
        {"name": "t", "params": None, "args": [b]},
        {"name": "t", "params": None, "args": [c]},
        {"name": "h", "params": None, "args": [c]},
        {**cx_common_entries, "args": [a, b]},
        {"name": "t", "params": None, "args": [a]},
        {"name": "tdg", "params": None, "args": [b]},
        {**cx_common_entries, "args": [a, b]},
    ]
    return subgates


def cswap_macro(a, b, c):
    """
    Macro for cswap gate.

    For internal use by :mod: `~dqc_simulator.software.ast2dqc_circuit`.
    Further compilation will be needed before the output can be interpretted
    by the simulator itself.

    Parameters
    ----------
    a : int
        Control qubit index
    b : int
        Index of first qubit to be swapped
    c : int
        Index of second qubit to be swapped

    Returns
    -------
    subgates : list of dicts
        The gates comprising the macro.
    """
    cx_gate = {"name": "cx", "params": None, "args": [c, b]}
    subgates = [cx_gate, *ccx_macro(a, b, c), cx_gate]
    return subgates


def crx_macro(lambda_var, a, b):
    """
    Macro for CRX gate.

    For internal use by :mod: `~dqc_simulator.software.ast2dqc_circuit`.
    Further compilation will be needed before the output can be interpretted
    by the simulator itself.

    Parameters
    ----------
    lambda_var : float
        The rotation angle
    a : int
        Control qubit index
    b : int
        Target qubit index

    Returns
    -------
    subgates : list of dicts
        The gates comprising the macro.

    References
    ----------
    See [1]_ for further info on lambda_var in the context of the CRZ gate.

    .. [1] https://arxiv.org/abs/1707.03429
    """
    subgates = [
        {"name": "u1", "params": ["pi/2"], "args": [b]},
        {"name": "cx", "params": None, "args": [a, b]},
        {"name": "u3", "params": [minus + lambda_var + over2, "0", "0"], "args": [b]},
        {"name": "cx", "params": None, "args": [a, b]},
        {"name": "u3", "params": [lambda_var + "/2", "-pi/2", "0"], "args": [b]},
    ]
    return subgates


def cry_macro(lambda_var, a, b):
    """
    Macro for CRY gate.

    For internal use by :mod: `~dqc_simulator.software.ast2dqc_circuit`.
    Further compilation will be needed before the output can be interpretted
    by the simulator itself.

    Parameters
    ----------
    lambda_var : float
        The rotation angle
    a : int
        Control qubit index
    b : int
        Target qubit index

    Returns
    -------
    subgates : list of dicts
        The gates comprising the macro.

    References
    ----------
    See [1]_ for further info on lambda_var in the context of the CRZ gate.

    .. [1] https://arxiv.org/abs/1707.03429
    """
    cx_gate = {"name": "cx", "params": None, "args": [a, b]}
    subgates = [
        {"name": "ry", "params": [lambda_var + over2], "args": [b]},
        cx_gate,
        {"name": "ry", "params": [minus + lambda_var + over2], "args": [b]},
        cx_gate,
    ]
    return subgates


def crz_macro(lambda_var, a, b):
    """
    Macro for CRX gate.

    For internal use by :mod: `~dqc_simulator.software.ast2dqc_circuit`.
    Further compilation will be needed before the output can be interpretted
    by the simulator itself.

    Parameters
    ----------
    lambda_var : float
        The rotation angle
    a : int
        Control qubit index
    b : int
        Target qubit index

    Returns
    -------
    subgates : list of dicts
        The gates comprising the macro.

    References
    ----------
    See [1]_ for further info on lambda_var in the context of the CRZ gate.

    .. [1] https://arxiv.org/abs/1707.03429
    """
    cx_gate = {"name": "cx", "params": None, "args": [a, b]}
    subgates = [
        {"name": "rz", "params": [lambda_var + over2], "args": [b]},
        cx_gate,
        {"name": "rz", "params": [minus + lambda_var + over2], "args": [b]},
        cx_gate,
    ]
    return subgates


def cu1_macro(lambda_var, a, b):
    """
    Macro for CU1 gate defined in IBM quantum experience standard header.

    For internal use by :mod: `~dqc_simulator.software.ast2dqc_circuit`.
    Further compilation will be needed before the output can be interpretted
    by the simulator itself.

    Parameters
    ----------
    lambda_var : float
        The phase applied.
    a : int
        Control qubit index
    b : int
        Target qubit index

    Returns
    -------
    subgates : list of dicts
        The gates comprising the macro.

    References
    ----------
    See [1]_ for further info on lambda_var.

    .. [1] https://arxiv.org/abs/1707.03429
    """
    cx_gate = {"name": "cx", "params": None, "args": [a, b]}
    subgates = [
        {"name": "u1", "params": [lambda_var + over2], "args": [b]},
        cx_gate,
        {"name": "u1", "params": [minus + lambda_var + over2], "args": [b]},
        cx_gate,
        {"name": "u1", "params": [lambda_var + over2], "args": [b]},
    ]
    return subgates


def cp_macro(lambda_var, a, b):
    """
    Macro for CP(lambda_var) gate (alias of CU1 gate).

    For internal use by :mod: `~dqc_simulator.software.ast2dqc_circuit`.
    Further compilation will be needed before the output can be interpretted
    by the simulator itself.

    Parameters
    ----------
    lambda_var : float
        The phase.
    a : int
        Control qubit index
    b : int
        Target qubit index

    Returns
    -------
    subgates : list of dicts
        The gates comprising the macro.

    References
    ----------
    See [1]_ for adiscussion of lambda_var.

    .. [1] https://arxiv.org/abs/1707.03429
    """
    cx_gate = {"name": "cx", "params": None, "args": [a, b]}
    subgates = [
        {"name": "p", "params": [lambda_var + over2], "args": [b]},
        cx_gate,
        {"name": "p", "params": [minus + lambda_var + over2], "args": [b]},
        cx_gate,
        {"name": "p", "params": [lambda_var + over2], "args": [b]},
    ]
    return subgates


def cu3_macro(theta, phi, lambda_var, c, t):
    """
    Macro for CU3 gate defined by IBM standard header.

    For internal use by :mod: `~dqc_simulator.software.ast2dqc_circuit`.
    Further compilation will be needed before the output can be interpretted
    by the simulator itself.

    Parameters
    ----------
    theta : float
    lambda_var : float
    c : int
        Control qubit index
    t : int
        Target qubit index

    Returns
    -------
    subgates : list of dicts
        The gates comprising the macro.

    References
    ----------
    See [1]_ for adiscussion of theta, phi, and lambda_var.

    .. [1] https://arxiv.org/abs/1707.03429
    """
    subgates = [
        {"name": "u1", "params": [lpar + lambda_var + plus + phi + ")/2"], "args": [c]},
        {"name": "u1", "params": [lpar + lambda_var + "-" + phi + ")/2"], "args": [t]},
        {"name": "cx", "params": None, "args": [c, t]},
        {
            "name": "u3",
            "params": ["-" + theta + "/2", "0", "-(" + phi + "+" + lambda_var + ")/2"],
            "args": [t],
        },
        {"name": "cx", "params": None, "args": [c, t]},
        {"name": "u3", "params": [theta + "/2", phi, "0"], "args": [t]},
    ]
    return subgates


def csx_macro(a, b):
    """
    Macro for CSX gate.

    For internal use by :mod: `~dqc_simulator.software.ast2dqc_circuit`.
    Further compilation will be needed before the output can be interpretted
    by the simulator itself.

    Parameters
    ----------
    a : int
        Control qubit index
    b : int
        Target qubit index

    Returns
    -------
    subgates : list of dicts
        The gates comprising the macro.
    """
    h_gate = {"name": "h", "params": None, "args": [b]}
    subgates = [h_gate, *cu1_macro("pi/2", a, b), h_gate]
    return subgates


def cu_macro(theta, phi, lambda_var, gamma, c, t):
    """
    Macro for CU(theta, phi, lambda_var) gate.

    For internal use by :mod: `~dqc_simulator.software.ast2dqc_circuit`.
    Further compilation will be needed before the output can be interpretted
    by the simulator itself.

    Parameters
    ----------
    theta : float
    phi : float
    lambda_var : float
    c : int
        Control qubit index
    t : int
        Target qubit index

    Returns
    -------
    subgates : list of dicts
        The gates comprising the macro.

    References
    ----------
    See [1]_ for adiscussion of theta, phi, and lambda_var.

    .. [1] https://arxiv.org/abs/1707.03429
    """
    subgates = [
        {"name": "p", "params": [gamma], "args": [c]},
        {
            "name": "p",
            "params": [lpar + lambda_var + plus + phi + rpar + over2],
            "args": [c],
        },
        {
            "name": "p",
            "params": [lpar + lambda_var + minus + phi + rpar + over2],
            "args": [t],
        },
        {"name": "cx", "params": None, "args": [c, t]},
        {
            "name": "u",
            "params": [
                minus + theta + over2,
                "0",
                minus + lpar + lambda_var + rpar + over2,
            ],
            "args": [t],
        },
        {"name": "cx", "params": None, "args": [c, t]},
        {"name": "u", "params": [theta + over2, phi, "0"], "args": [t]},
    ]
    return subgates


def rxx_macro(theta, a, b):
    """
    Macro for RXX(theta) gate.

    Parameters
    ----------
    theta : float
    a : int
        First qubit index
    b : int
        Second qubit index

    subgates : list of dicts
        The gates comprising the macro.

    References
    ----------
    See [1]_ for adiscussion of theta.

    .. [1] https://arxiv.org/abs/1707.03429
    """
    h_gate = {"name": "h", "params": None, "args": [b]}
    cx_gate = {"name": "cx", "params": None, "args": [a, b]}
    subgates = [
        {"name": "u3", "params": ["pi/2", theta, "0"], "args": [a]},
        h_gate,
        cx_gate,
        {"name": "u1", "params": ["-" + theta], "args": [b]},
        cx_gate,
        h_gate,
        {"name": "u2", "params": ["-pi", "pi-" + theta], "args": [a]},
    ]
    return subgates


def rzz_macro(theta, a, b):
    """
    Macro for RZZ(theta) gate

    Parameters
    ----------
    theta : float
        The rotation angle
    a, b : int
        The index of the first and second qubits, respectively.

    Returns
    -------
    subgates : list of dicts
        The gates comprising the macro.

    References
    ----------
    See [1]_ for adiscussion of theta.

    .. [1] https://arxiv.org/abs/1707.03429
    """
    cx_gate = {"name": "cx", "params": None, "args": [a, b]}
    subgates = [cx_gate, {"name": "u1", "params": [theta], "args": [b]}, cx_gate]
    return subgates


def rccx_macro(a, b, c):
    """
    Macro for RCCX gate

    Parameters
    ----------
    a, b : int
        The indices of the control qubit
    c : int
        the index of the target qubit

    Returns
    -------
    subgates : list of dicts
        The gates comprising the macro.

    """
    subgates = [
        {"name": "u2", "params": ["0", "pi"], "args": [c]},
        {"name": "u1", "params": ["pi/4"], "args": [c]},
        {"name": "cx", "params": None, "args": [b, c]},
        {"name": "u1", "params": ["-pi/4"], "args": [c]},
        {"name": "cx", "params": None, "args": [a, c]},
        {"name": "u1", "params": ["pi/4"], "args": [c]},
        {"name": "cx", "params": None, "args": [b, c]},
        {"name": "u1", "params": ["-pi/4"], "args": [c]},
        {"name": "u2", "params": ["0", "pi"], "args": [c]},
    ]
    return subgates


def rc3x_macro(a, b, c, d):
    """
    Macro for RCCCX gate.

    Parameters
    ----------
    a, b, c : int
        The indices of the control qubits
    d : int
        Index of the target qubit

    Returns
    -------
    subgates : list of dicts
        The gates comprising the macro.
    """
    subgates = [
        {"name": "u2", "params": ["0", "pi"], "args": [d]},
        {"name": "u1", "params": ["pi/4"], "args": [d]},
        {"name": "cx", "params": None, "args": [c, d]},
        {"name": "u1", "params": ["-pi/4"], "args": [d]},
        {"name": "u2", "params": ["0", "pi"], "args": [d]},
        {"name": "cx", "params": None, "args": [a, d]},
        {"name": "u1", "params": ["pi/4"], "args": [d]},
        {"name": "cx", "params": None, "args": [b, d]},
        {"name": "u1", "params": ["-pi/4"], "args": [d]},
        {"name": "cx", "params": None, "args": [a, d]},
        {"name": "u1", "params": ["pi/4"], "args": [d]},
        {"name": "cx", "params": None, "args": [b, d]},
        {"name": "u1", "params": ["-pi/4"], "args": [d]},
        {"name": "u2", "params": ["0", "pi"], "args": [d]},
        {"name": "u1", "params": ["pi/4"], "args": [d]},
        {"name": "cx", "params": None, "args": [c, d]},
        {"name": "u1", "params": ["-pi/4"], "args": [d]},
        {"name": "u2", "params": ["0", "pi"], "args": [d]},
    ]
    return subgates


def c3x_macro(a, b, c, d):
    """
    Macro for CCCX gate.

    Parameters
    ----------
    a, b, c : int
        The indices of the control qubits
    d : int
        Index of the target qubit

    Returns
    -------
    subgates : list of dicts
        The gates comprising the macro.
    """
    subgates = [
        {"name": "h", "params": None, "args": [d]},
        {"name": "p", "params": ["pi/8"], "args": [a]},
        {"name": "p", "params": ["pi/8"], "args": [b]},
        {"name": "p", "params": ["pi/8"], "args": [c]},
        {"name": "p", "params": ["pi/8"], "args": [d]},
        {"name": "cx", "params": None, "args": [a, b]},
        {"name": "p", "params": ["-pi/8"], "args": [b]},
        {"name": "cx", "params": None, "args": [a, b]},
        {"name": "cx", "params": None, "args": [b, c]},
        {"name": "p", "params": ["-pi/8"], "args": [c]},
        {"name": "cx", "params": None, "args": [a, c]},
        {"name": "p", "params": ["pi/8"], "args": [c]},
        {"name": "cx", "params": None, "args": [b, c]},
        {"name": "p", "params": ["-pi/8"], "args": [c]},
        {"name": "cx", "params": None, "args": [a, c]},
        {"name": "cx", "params": None, "args": [c, d]},
        {"name": "p", "params": ["-pi/8"], "args": [d]},
        {"name": "cx", "params": None, "args": [b, d]},
        {"name": "p", "params": ["pi/8"], "args": [d]},
        {"name": "cx", "params": None, "args": [c, d]},
        {"name": "p", "params": ["-pi/8"], "args": [d]},
        {"name": "cx", "params": None, "args": [a, d]},
        {"name": "p", "params": ["pi/8"], "args": [d]},
        {"name": "cx", "params": None, "args": [c, d]},
        {"name": "p", "params": ["-pi/8"], "args": [d]},
        {"name": "cx", "params": None, "args": [b, d]},
        {"name": "p", "params": ["pi/8"], "args": [d]},
        {"name": "cx", "params": None, "args": [c, d]},
        {"name": "p", "params": ["-pi/8"], "args": [d]},
        {"name": "cx", "params": None, "args": [a, d]},
        {"name": "h", "params": None, "args": [d]},
    ]
    return subgates


def c3sqrtx_macro(a, b, c, d):
    """
    Macro for RCCCSqrt(X) gate.

    Parameters
    ----------
    a, b, c : int
        The indices of the control qubits
    d : int
        Index of the target qubit

    Returns
    -------
    subgates : list of dicts
        The gates comprising the macro.
    """
    h_gate = {"name": "h", "params": None, "args": [d]}
    subgates = [
        h_gate,
        *cu1_macro("pi/8", a, d),
        h_gate,
        {"name": "cx", "params": None, "args": [a, b]},
        h_gate,
        *cu1_macro("-pi/8", b, d),
        h_gate,
        {"name": "cx", "params": None, "args": [a, b]},
        h_gate,
        *cu1_macro("pi/8", b, d),
        h_gate,
        {"name": "cx", "params": None, "args": [b, c]},
        h_gate,
        *cu1_macro("-pi/8", c, d),
        h_gate,
        {"name": "cx", "params": None, "args": [a, c]},
        h_gate,
        *cu1_macro("pi/8", c, d),
        h_gate,
        {"name": "cx", "params": None, "args": [b, c]},
        h_gate,
        *cu1_macro("-pi/8", c, d),
        h_gate,
        {"name": "cx", "params": None, "args": [a, c]},
        h_gate,
        *cu1_macro("pi/8", c, d),
        h_gate,
    ]
    return subgates


def c4x_macro(a, b, c, d, e):
    """
    Macro for CCCCX gate.

    Parameters
    ----------
    a, b, c, d : int
        The indices of the control qubits
    e : int
        Index of the target qubit

    Returns
    -------
    subgates : list of dicts
        The gates comprising the macro.
    """
    h_gate = {"name": "h", "params": None, "args": [e]}
    subgates = [
        h_gate,
        *cu1_macro("pi/2", d, e),
        h_gate,
        *c3x_macro(a, b, c, d),
        h_gate,
        *cu1_macro("-pi/2", d, e),
        h_gate,
        *c3x_macro(a, b, c, d),
        *c3sqrtx_macro(a, b, c, e),
    ]
    return subgates



# -*- coding: utf-8 -*-
# =============================================================================
# Created on Wed Sep 20 12:13:12 2023
#
# @author: kenny
# =============================================================================
"""
Additional quantum gates to those included natively in NetSquid.

The gates defined in this module are
:class:`~netsquid.components.instructions.Instruction` objects or tuples
containing :class:`~netsquid.components.instructions.Instruction` objects or
functions returning :class:`~netsquid.components.instructions.Instruction`
 objects or tuples containing instructions and
:class:`~netsquid.qubits.operators.Operator` objects to be used by those
instructions. In short, the gates in this module are intended for backend use
and will utilised directly by the simulation.
"""

# This script defines additional quantum gates to those included natively in
# NetSquid (as of version 1.1.6).

import numpy as np

from netsquid.components import instructions as instr
from netsquid.qubits.operators import Operator, H, T, I, S, Y, create_rotation_op


def make_state_gen_op(alpha, beta):
    """
    For making operators that generate arbitrary, pure, single-qubit states.

    Parameters
    ----------
    alpha, beta : float
        The coefficients of |0> and |1>, respectively, in the state to
        generate.

    Returns
    -------
    state_gen_op : :class:`~netsquid.qubits.operators.Operator`
        An operator describing a fictitious quantum gate which can create an
        arbitrary single qubit gate.
    """
    state_gen_op = Operator(
        "state_generating_op",
        np.array([[alpha, 0], [0, beta]]) @ np.array([[1, 1], [1, -1]]),
    )
    return state_gen_op


def INSTR_ARB_GEN(alpha, beta):
    """
    For generating arbitrary quantum states.

    .. deprecated::
    This function will be deprecated in a future version. It was implemented
    to allow arbitrary quantum states to be generated, however, this is now
    done using :func: `~dqc_simulator.qlib.make_state_gen_op`
    and an instruction :class:`~netsquid.components.instructions.Instruction`
    without an operator in the future. This will avoid the need to bake
    `alpha` and `beta` into the
    :class:`~netsquid.components.qprocessor.QuantumProcessor`.

    Creates :class:`~netsquid.components.instructions.Instruction` that
    generates arbitrary quantum states.

    Parameters
    ----------
    alpha, beta : float
        The coefficients of |0> and |1>, respectively, in the state to
        generate.

    Raises
    ------
    ValueError
        Alerts user to erroneous choice of `alpha` and `beta`, which would
        generate an unnormalised quantum state.

    Returns
    -------
    instruction : :class:`~netsquid.components.instructions.Instruction`
        An instruction that creates arbitrary quantum states.

    """
    # THIS WILL BE DEPRECATED IN FUTURE VERSION
    total_probability_q1 = abs(alpha) ** 2 + abs(beta) ** 2
    if round(total_probability_q1, 3) != 1.000:
        raise ValueError("alpha and beta do not give normalised input to circuit")

    state_gen_op = make_state_gen_op(alpha, beta)
    instruction = instr.IGate("state_gen_gate", state_gen_op)
    return instruction


INSTR_CH = instr.IGate("CH", H.ctrl)  # creates CH gate as instruction

INSTR_CT = instr.IGate("CT", T.ctrl)

INSTR_IDENTITY = instr.IGate("I", operator=I)

INSTR_T_DAGGER = instr.IGate("T_dagger", operator=T.conj)

INSTR_S_DAGGER = instr.IGate("S_dagger", operator=S.conj)


# instructions without operations. The operation will be specified later.
# These instructions should be used when we want to define a PhysicalInstruction
# for an operation dependent on continuous paramters.
INSTR_SINGLE_QUBIT_UNITARY = instr.IGate("single_qubit_unitary", num_positions=1)
INSTR_TWO_QUBIT_UNITARY = instr.IGate("two_qubit_unitary", num_positions=2)
INSTR_SINGLE_QUBIT_NEGLIGIBLE_TIME = instr.IGate(
    "neglibible_time_instr", num_positions=1
)
INSTR_TWO_QUBIT_NEGLIGIBLE_TIME = instr.IGate("neglible_time_2_qubit", num_positions=2)


def INSTR_U(theta, phi, lambda_var, controlled=False):
    """
    (controlled) single qubit unitary in sim backend-readable format.

    Parameters
    ----------
    theta : float
        An angle in radians.
    phi : float
        An angle in radians.
    lambda_var : float
        An angle in radians.
    controlled : bool
        Whether the gate should be controlled (ie, a CU gate is applied rather
        than a U gate) or not. Default is False.

    Returns
    -------
    instructionNop : tuple
        The NetSquid :class:`~netsquid.components.instructions.Instruction`
        and operation :class:`~netsquid.qubits.operators.Operator`
        needed to carry out this gate.
        Here, the instruction essentially indicates all relevant metadata and
        can be used to dictate which noise should be applied to the gate,
        while the operator is the matrix used to apply the actual quantum
        operation defining the gate.
    """
    a11 = np.exp(-1j * (phi + lambda_var) / 2) * np.cos(theta / 2)
    a12 = -np.exp(-1j * (phi - lambda_var) / 2) * np.sin(theta / 2)
    a21 = np.exp(1j * (phi - lambda_var) / 2) * np.sin(theta / 2)
    a22 = np.exp(1j * (phi + lambda_var) / 2) * np.cos(theta / 2)
    op = Operator("single_qubit_unitary_op", np.array([[a11, a12], [a21, a22]]))
    instruction = INSTR_SINGLE_QUBIT_UNITARY
    if controlled:
        op = op.ctrl
        instruction = INSTR_TWO_QUBIT_UNITARY
    elif not isinstance(controlled, bool):
        raise TypeError(f"{controlled} is not of type `bool' ")
    instructionNop = (instruction, op)
    return instructionNop


INSTR_CY = instr.IGate("CY", operator=Y.ctrl)


def _add_options2single_qubit_gate(op, controlled, conjugate):
    instruction = INSTR_SINGLE_QUBIT_UNITARY
    if controlled:
        op = op.ctrl
        instruction = INSTR_TWO_QUBIT_UNITARY
    elif not isinstance(controlled, bool):
        raise TypeError(
            f"The value for controlled ({controlled}) is not of type `bool' "
        )

    if conjugate:
        op = op.conj
    elif not isinstance(conjugate, bool):
        raise TypeError(f"The value for conjugate ({conjugate}) is not of type `bool' ")

    return (instruction, op)


def instrNop_RZ(angle, controlled=False, conjugate=False):
    """
    Defines RZ gate in sim backend-readable format.

    Parameters
    ----------
    angle : float
        Angle of rotation in radians.
    controlled : bool, optional
        Whether this is implemented as a control gate (CRz) or not. The default
        is False.
    conjugate : bool, optional
        Whether to implement the complex conjugate of Rz or not. The default
        is False.

    Returns
    -------
    instructionNop : tuple
        The NetSquid :class:`~netsquid.components.instructions.Instruction`
        and operation :class:`~netsquid.qubits.operators.Operator`
        needed to carry out this gate.
        Here, the instruction essentially indicates all relevant metadata and
        can be used to dictate which noise should be applied to the gate,
        while the operator is the matrix used to apply the actual quantum
        operation defining the gate.

    """
    op = create_rotation_op(angle, rotation_axis=(0, 0, 1))
    instructionNop = _add_options2single_qubit_gate(op, controlled, conjugate)
    return instructionNop


instrNop_SX = (
    INSTR_SINGLE_QUBIT_UNITARY,
    Operator("SX", S.conj.arr @ H.arr @ S.conj.arr),
)

instrNop_SXDG = (INSTR_SINGLE_QUBIT_UNITARY, Operator("SXDG", S.arr @ H.arr @ S.arr))

instrNop_CSX = (INSTR_TWO_QUBIT_UNITARY, instrNop_SX[1].ctrl)



# -*- coding: utf-8 -*-
"""
Tools for the rough parsing of an abstract syntax tree.

This module defines classes that are used to parse an abstract syntax tree
created using :func: `~dqc_simulator.software.qasm2ast.qasm2ast`, which is
adapted from the open source nuqasm2 package. It heavily uses the openQASM 2.0
grammar. The abstract syntax tree is parsed into a
:class:`~dqc_simulator.software.dqc_circuit.DqcCircuit`.

Notes
-----
This module uses open source code written by pyparsing module author Paul
McGuire
(https://github.com/pyparsing/pyparsing/blob/master/examples/fourFn.py)
in accordance with the MIT license. This is marked with an appropriate comment
or in the docstring notes at the relevant points in the script.

A key difference between my code and a proper QISKIT parser is that I implement
 the standard library gates primarily as built in gates (ie, as literals)
 rather than building them from openQASM’s two native gate sets as macros whose
 expansion is deferred until runtime (which is what is done by openQASM 2.0).
 This means that all of my gates are basically implemented as native gates.
 This aligns with NetSquid’s platform agnostic philosophy. I feel that any
 breaking down of gates into subroutines, should be done for specific platforms
 and included in compilation (not just at runtime). All of this means I am not
 strictly parsing openQASM 2.0 but rather directly converting it to something
 else.

"""

import abc
import operator
import warnings

import numpy as np
import pyparsing as pp
from netsquid.components import instructions as instr

from dqc_simulator.software.dqc_circuit import DqcCircuit
from dqc_simulator.software.qasm2ast import ASTType, QasmParsingElement
from dqc_simulator.qlib import gates
from dqc_simulator.qlib import macros4parsing as macros


class ExpQasm(QasmParsingElement):
    """
    Used to interpret mathematical expressions.

    Defines <exp> non-terminal defined in openQASM 2.0 grammar.

    Notes
    -----
    This is slightly modified from code written by pyparsing module author Paul
    McGuire
    (https://github.com/pyparsing/pyparsing/blob/master/examples/fourFn.py)
    and used in accordance with the MIT license:
    """

    def __init__(self):
        self.exp_qasmStack = []

    def _push_first(self, toks):
        self.exp_qasmStack.append(toks[0])

    def _push_unary_minus(self, toks):
        for t in toks:
            if t == "-":
                self.exp_qasmStack.append("unary -")
            else:
                break

    def _insert_fn_argcount_tuple(self, t):
        fn = t.pop(0)
        num_args = len(t[0])
        t.insert(0, (fn, num_args))

    def define_grammar(self):
        pi = pp.CaselessKeyword("pi")
        # fnumber = Combine(Word("+-"+nums, nums) +
        #                    Optional("." + Optional(Word(nums))) +
        #                    Optional(e + Word("+-"+nums, nums)))
        # or use provided pyparsing_common.number, but convert back to str:
        # fnumber = ppc.number().addParseAction(lambda t: str(t[0]))
        fnumber = pp.Regex(r"[+-]?\d+(?:\.\d*)?(?:[eE][+-]?\d+)?")
        ident = pp.Word(pp.alphas, pp.alphanums + "_$")

        plus, minus, mult, div = map(pp.Literal, "+-*/")
        addop = plus | minus
        multop = mult | div
        expop = pp.Literal("^")

        exp_qasm = pp.Forward()

        fn_call = (ident + self.lpar - exp_qasm + self.rpar).setParseAction(
            self._insert_fn_argcount_tuple
        )
        atom = (
            addop[...]
            + (
                (fn_call | pi | fnumber | ident).setParseAction(self._push_first)
                | pp.Group(self.lpar + exp_qasm + self.rpar)
            )
        ).setParseAction(self._push_unary_minus)

        # by defining exponentiation as "atom [ ^ factor ]..." instead of
        # "atom [ ^ atom ]...", we get right-to-left
        # exponents, instead of left-to-right that is, 2^3^2 = 2^(3^2), not (2^3)^2.
        factor = pp.Forward()
        factor <<= atom + (expop + factor).setParseAction(self._push_first)[...]
        term = factor + (multop + factor).setParseAction(self._push_first)[...]
        exp_qasm <<= term + (addop + term).setParseAction(self._push_first)[...]
        return exp_qasm

    # END of code modified from
    # https://github.com/pyparsing/pyparsing/blob/master/examples/fourFn.py


# =============================================================================
#     arith_expr = pp.infix_notation(exp_qasm, arith_ops, '(', ')')
#     unary_op_acting = pp.Group(unaryop + lpar + exp_qasm + rpar)
# # =============================================================================
# #     atom = arith_expr | unary_op_acting | number
# # =============================================================================
#     atom = number | arith_expr | unary_op_acting
#     exp_qasm <<= atom
# =============================================================================
# =============================================================================
#     exp_qasm = unary_op_acting ^ expr_or_number
# =============================================================================
# facilitating arithmatic combinations of exp_qasm terms:
# =============================================================================
#     exp_qasm = pp.infix_notation(exp_qasm, arith_ops, '(', ')')
# =============================================================================


class NonTerminalInterpreter(QasmParsingElement, metaclass=abc.ABCMeta):
    """Abstract base class for a non-terminal interpreter"""

    @abc.abstractmethod
    def interpret(self, unparsed_terminal_token):
        """
        Should be overwritten with a method interpretting an unparsed string
        corresponding to a terminal token. The terminal token should be of a
        specific non-terminal type, which should be indicated in the subclass
        name.

        Parameters
        ----------
        unparsed_terminal_token : str
            An unparsed string corresponding to a terminal token. The terminal
            token should be of a specific non-terminal type, which should be
            indicated in the subclass name.

        Raises
        ------
        NotImplementedError
            Error for if this method is not overwritten.
        """
        raise NotImplementedError(
            "This method should be overwritten when defining a subclass."
        )


class ExpQasmInterpreter(NonTerminalInterpreter):
    """
    Interprets mathematical expressions in string form as python objects.

    Attributes
    ----------
    unaryop_funcs : dict
        Lookup table of unary functions relating strings with the function
        name to the corresponding from `numpy` objects
    arith_op_funcs : dict
        Lookup table relating strings containing symbols to basic arithmetic
        operators.
    """

    def __init__(self):
        pass  # just defining the class instance as self

    unaryop_funcs = {
        "sin": np.sin,
        "cos": np.cos,
        "tan": np.tan,
        "exp": np.exp,
        "ln": np.log,
        "sqrt": np.sqrt,
    }

    arith_op_funcs = opn = {
        "+": operator.add,
        "-": operator.sub,
        "*": operator.mul,
        "/": operator.truediv,
        "^": operator.pow,
    }

    # =============================================================================
    #     def _do_nothing(self):
    #         """A function that does the identity transformation (that is to say
    #         nothing)"""
    #         pass
    # =============================================================================

    def _evaluate_stack(self, s):
        # modifying code written by pyparsing module author Paul McGuire
        # (https://github.com/pyparsing/pyparsing/blob/master/examples/fourFn.py)
        # in accordance with the MIT license:
        op, num_args = s.pop(), 0
        if isinstance(op, tuple):
            op, num_args = op
        if op == "unary -":
            return -self._evaluate_stack(s)
        if op in self.arith_op_funcs:
            # note: operands are pushed onto the stack in reverse order
            op2 = self._evaluate_stack(s)
            op1 = self._evaluate_stack(s)
            return self.arith_op_funcs[op](op1, op2)
        elif op == "pi":
            return np.pi  # 3.1415926535
        elif op in self.unaryop_funcs:
            # note: args are pushed onto the stack in reverse order
            args = reversed([self._evaluate_stack(s) for _ in range(num_args)])
            return self.unaryop_funcs[op](*args)
        elif op[0].isalpha():
            raise Exception("invalid identifier '%s'" % op)
        else:
            return float(op)
        # END of code modified from
        # https://github.com/pyparsing/pyparsing/blob/master/examples/fourFn.py

    def interpret(self, exp_qasm_string):
        exp_qasm = ExpQasm()
        exp_qasm_grammar = exp_qasm.define_grammar()
        # The 'ParseAction's a added to the parse_string method in what follows
        # cause it to update the exp_qasmStack
        exp_qasm_grammar.parse_string(exp_qasm_string, parseAll=True)
        return self._evaluate_stack(exp_qasm.exp_qasmStack[:])


class ArgumentInterpreter(NonTerminalInterpreter):
    def parse_argument_non_terminal(self, argument_terminal):
        """
        Tokenizes the <argument> non-terminal from the grammar specified in
        the openQASM 2.0 paper (https://arxiv.org/abs/1707.03429)

        Parameters
        ----------
        argument_terminal : str
            A string contain some terminal token which is an example of the
            <argument> non-terminal defined in the openQASM 2.0 paper
            (https://arxiv.org/abs/1707.03429).

        Returns
        ------
        parsed_arg : list
            The parsed argument. It contains the metadata indicating the type
            of argument ('qubit' or 'reg'), the name of the quantum register
            as a str (reg_name) and the bit index (iff the type of argument is
            'qubit')
        """
        if "[" in argument_terminal:  # is (qu)bit
            arg_type = "qubit"
            split_bit = argument_terminal.replace("]", "").split("[")
            reg_name = split_bit[0]
            bit_index = split_bit[1]
            parsed_arg = [arg_type, reg_name, bit_index]
        elif "[" not in argument_terminal:
            arg_type = "reg"
            reg_name = argument_terminal
            parsed_arg = [arg_type, reg_name]
        return parsed_arg

    def interpret(self, argument_terminal):
        parsed_arg = self.parse_argument_non_terminal(argument_terminal)
        arg_type = parsed_arg[0]
        if arg_type == "qubit":
            parsed_arg[-1] = int(parsed_arg[-1])
        interpretted_arg = parsed_arg
        return interpretted_arg


class Ast2SimReadable(metaclass=abc.ABCMeta):
    """Abstract base class for various converters to simulation readable
    commands

    Parameters
    ----------
    ast_c_sect_element : dict
        An element of the AST's 'c_sect'. This typically corresponds to a
        parsed line of .qasm source code but in some rare instances it
        could be multiple lines

    dqc_circuit : instance of DqcCircuit defined above
        The specs need to make a circuit that could run on a DQC (it may
        actually be monolithic but is written in a DQC suitable form)
    """

    def __init__(self, ast_c_sect_element, dqc_circuit):
        self.ast_c_sect_element = ast_c_sect_element
        self.dqc_circuit = dqc_circuit

    @abc.abstractmethod
    def make_sim_readable(self):
        """
        Should be overwritten with method taking same arguments which translates
        the qasm terminal into something intelligible by the target code.

        Parameters
        ----------
        terminal : str
            A QASM terminal symbol.

        Raises
        -------
            `NotImplementedError`
        """
        raise NotImplementedError


class AstUnknown(Ast2SimReadable):
    """For handling unknown cases"""

    def make_sim_readable(self):
        """
        Raises
        ------
        ValueError
            Error due to unrecognised element in the abstract syntax tree
        """
        raise ValueError(
            f"Unknown element {self.ast_c_sect_element} "
            "identified while parsing. There may have been an"
            " error the .qasm file or a failure to recognise the "
            "issue when parsing."
        )


class AstComment(Ast2SimReadable):
    """Interprets comments in AST"""

    def make_sim_readable(self):
        """
        Returns
        -------
        :class:`~dqc_simulator.software.dqc_circuit.DqcCircuit`
            The unchanged DQC circuit.

        """
        return self.dqc_circuit  # This ignores comments in the source code,
        # as they are not needed in the gate_list to be interpreted


class AstQreg(Ast2SimReadable):
    """Interprets qreg commands found in AST."""

    def make_sim_readable(self):
        """
        Returns
        -------
        :class:`~dqc_simulator.software.dqc_circuit.DqcCircuit`
            The updated DQC circuit.
        """
        qreg_name = self.ast_c_sect_element["qreg_name"]
        qreg_size = self.ast_c_sect_element["qreg_num"]
        self.dqc_circuit.qregs[qreg_name] = {
            "size": int(qreg_size),
            "starting_index": self.dqc_circuit.qubit_count,
        }
        self.dqc_circuit.qubit_count = self.dqc_circuit.qubit_count + int(qreg_size)
        return self.dqc_circuit


class AstCreg(Ast2SimReadable):
    """Interprets creg commands found in AST."""

    def make_sim_readable(self):
        """
        Returns
        -------
        :class:`~dqc_simulator.software.dqc_circuit.DqcCircuit`
            The updated DQC circuit.
        """
        creg_name = self.ast_c_sect_element["creg_name"]
        creg_number = self.ast_c_sect_element["creg_num"]
        self.dqc_circuit.cregs[creg_name] = int(creg_number)
        # TO DO: maybe make more like AstQreg and have both size and starting
        # index
        return self.dqc_circuit


class AstMeasure(Ast2SimReadable):
    """Interprets measure commands found in AST."""

    def make_sim_readable(self):
        """
        Returns
        -------
        :class:`~dqc_simulator.software.dqc_circuit.DqcCircuit`
            An unchanged DQC circuit. Right now, measurements specified in
            the abstract syntax tree are ignored.
        """
        warnings.warn("WARNING: measurement present in .qasm code but ignored")
        return self.dqc_circuit  # FOR NOW THIS DOES NOTHING AND MEASUREMENTS
        # ARE SIMPLY IGNORED. I would like to make this
        # optional in the future, so that I can analyse
        # states at the end of circuits with many
        # measurements at the end automatically. That
        # said, it may be better to just change the
        # .qasm source code.


# =============================================================================
#         source_reg = self.ast_c_sect_element['source_reg']
#         target_reg = self.ast_c_sect_element['target_reg']
#         #'source_reg' and 'target_reg' are the relevant keys.
#         source_reg_name, source_qubit_index = (
#             self.tokenize_argument_non_terminal(source_reg))
#         target_reg_name, target_qubit_index = (
#             self.tokenize_argument_non_terminal(target_reg))
# # =============================================================================
# #         if source_qubit_index is not None and target_qubit_index is not None:
# # # =============================================================================
# # #             self.dqc_circuit.ops = (instr.INSTR_MEASURE, )
# # # =============================================================================
# # =============================================================================
#         #FINISH!!!!!
# =============================================================================


class AstBarrier(Ast2SimReadable):
    """Interprets barrier commands found in AST"""

    def make_sim_readable(self):
        """
        Returns
        -------
        :class:`~dqc_simulator.software.dqc_circuit.DqcCircuit`
            An unchanged DQC circuit. Right now, the barrier command is
            ignored.
        """
        warnings.warn("WARNING: 'barrier' command present in .qasm code but ignored")
        return self.dqc_circuit  # For now, this does nothing and the barrier
        # command is simply ignored.


class AstGate(Ast2SimReadable):
    """Interprets definitions and calls of quantum gates."""

    def _add_gate_spec_with_params2circuit(self, params, gate_name, gate_args):
        """
        Adds new gate specs.

        Adds new gate specifications to `self.dqc_circuit.ops`

        Parameters
        ----------
        params : list of float
            The interpretted numerical parameters used to specify the gate.
            Eg, for gate U(theta, phi, lambda) the parameters might be
            [1, 2.3, 1.2].
        gate_name : str
            The name of the gate.
        gate_args : list
            List of form: [qubit_index, qreg_name], where qubit_index is of
            type int and qreg_name of type str.
        """
        gate_spec = [self.dqc_circuit.native_gates[gate_name](*params), *gate_args]
        self.dqc_circuit.ops.append(gate_spec)

    def _add_gate_spec_without_params2circuit(self, gate_name, gate_args):
        """
        Adds new gate specs.

        Adds new gate specifications to self.dqc_circuit.ops

        Parameters
        ----------
        gate_name : str
            The name of the gate.
        gate_args : list
            List of form: [qubit_index, qreg_name], where qubit_index is of
            type int and qreg_name of type str.
        """
        gate_spec = [self.dqc_circuit.native_gates[gate_name], *gate_args]
        self.dqc_circuit.ops.append(gate_spec)

    def _add_gate_call2circuit(
        self, arg_interpreter, param_interpreter, params, args, gate_name
    ):
        interpreted_args = []
        for arg in args:
            interpreted_args.append(arg_interpreter(arg))

        interpreted_params = []
        if params:
            _add_gate_spec2circuit = self._add_gate_spec_with_params2circuit
            for param in params:
                interpreted_params.append(param_interpreter(param))
            gate_spec_elems = [interpreted_params, gate_name]

        elif params is None:
            gate_spec_elems = [gate_name]
            _add_gate_spec2circuit = self._add_gate_spec_without_params2circuit

        arg1 = interpreted_args[0]
        arg1_type = arg1[0]
        reg1_name = arg1[1]
        if len(interpreted_args) == 1:  # if single-qubit gate
            if arg1_type == "reg":
                reg_size = self.dqc_circuit.qregs[reg1_name]["size"]
                for qubit_index in range(reg_size):
                    gate_args = [qubit_index, reg1_name]
                    _add_gate_spec2circuit(*gate_spec_elems, gate_args)
            elif arg1_type == "qubit":
                qubit_index = arg1[2]
                gate_args = [qubit_index, reg1_name]
                _add_gate_spec2circuit(*gate_spec_elems, gate_args)

        elif len(interpreted_args) == 2:  # if two-qubit gate
            arg2 = interpreted_args[1]
            arg2_type = arg2[0]
            reg2_name = arg2[1]
            reg1_size = self.dqc_circuit.qregs[reg1_name]["size"]
            reg2_size = self.dqc_circuit.qregs[reg2_name]["size"]

            if arg1_type == arg2_type == "qubit":
                qubit_index1 = arg1[2]
                qubit_index2 = arg2[2]
                gate_args = [qubit_index1, reg1_name, qubit_index2, reg2_name]
                _add_gate_spec2circuit(*gate_spec_elems, gate_args)
            elif arg1_type == arg2_type == "reg":
                for qubit_index in range(reg1_size):
                    # if reg1_size != reg2_size there will be an error thrown
                    # later implicitly and so can just use reg1_size or
                    # reg2_size here, knowing regs MUST be same size. This is
                    # consistent with the openQasm 2.0 specification
                    gate_args = [qubit_index, reg1_name, qubit_index, reg2_name]
                    _add_gate_spec2circuit(*gate_spec_elems, gate_args)
            elif arg1_type == "qubit" and arg2_type == "reg":
                for qubit_index2 in range(reg2_size):
                    qubit_index1 = arg1[2]
                    gate_args = [qubit_index1, reg1_name, qubit_index2, reg2_name]
                    _add_gate_spec2circuit(*gate_spec_elems, gate_args)
            elif arg1_type == "reg" and arg2_type == "qubit":
                for qubit_index1 in range(reg1_size):
                    qubit_index2 = arg2[2]
                    gate_args = [qubit_index1, reg1_name, qubit_index2, reg2_name]
                    _add_gate_spec2circuit(*gate_spec_elems, gate_args)

    def make_sim_readable(self):
        """
        Returns
        -------
        :class:`~dqc_simulator.software.dqc_circuit.DqcCircuit`
            The updated DQC circuit.
        """
        arg_interpreter = ArgumentInterpreter().interpret
        param_interpreter = ExpQasmInterpreter().interpret
        params = self.ast_c_sect_element["param_list"]
        args = self.ast_c_sect_element["reg_list"]
        gate_name = self.ast_c_sect_element["op"]
        # define list of macros
        if gate_name in self.dqc_circuit.gate_macros:
            if params is not None:
                subgates = self.dqc_circuit.gate_macros[gate_name](*params, *args)
            # gate macros should be dict
            # of funcs which return expanded
            # macro, generated in InterpretGSect
            # maybe add recursive call here. The end goal is a list of subgates
            # where all subgates have been expanded fully in terms of native
            # gates.
            elif params is None:
                subgates = self.dqc_circuit.gate_macros[gate_name](*args)
            for subgate in subgates:
                self._add_gate_call2circuit(
                    arg_interpreter,
                    param_interpreter,
                    subgate["params"],
                    subgate["args"],
                    subgate["name"],
                )

        elif gate_name in self.dqc_circuit.native_gates:
            self._add_gate_call2circuit(
                arg_interpreter, param_interpreter, params, args, gate_name
            )

        return self.dqc_circuit


class AstCtl(Ast2SimReadable):
    def make_sim_readable(self):
        raise NotImplementedError("QASM if statements are not yet supported")


class AstCtl2(Ast2SimReadable):
    def make_sim_readable(self):
        raise NotImplementedError("QASM if statements are not yet supported")


class AstBlank(Ast2SimReadable):
    def make_sim_readable(self):
        raise NotImplementedError(
            "I am not quite sure what the Blank type in the parser is yet"
        )


class AstDeclaration_Qasm_2_0(Ast2SimReadable):
    """Interprets document preamble"""

    def make_sim_readable(self):
        """
        Returns
        -------
        :class:`~dqc_simulator.software.dqc_circuit.DqcCircuit`
            The updated DQC circuit.
        """
        return self.dqc_circuit  # do nothing. The error raised if this is not
        # the first line of uncommented code in the file should already
        # have been raised by the parser and so this line should do nothing
        # further


class AstInclude(Ast2SimReadable):
    """Interprets openQASM 2.0 'include' statements"""

    def make_sim_readable(self):
        """
        Returns
        -------
        :class:`~dqc_simulator.software.dqc_circuit.DqcCircuit`
            The updated DQC circuit.
        """
        # TO DO: exchange this all for 'return self.dqc_circuit' once the rest
        # of the code is ready to do without it. The include statement has
        # already been parsed into 'g_sect' when creating the ast and this
        # is what should be interpretted (not the ast['c_sect'])
        if self.ast_c_sect_element["include"] is None:
            return self.dqc_circuit
        elif self.ast_c_sect_element["include"] == "qelib1.inc":
            standard_lib_aliases4_native_gates = {
                "u3": gates.INSTR_U,  # alias of U native gate
                "u2": lambda phi, lambda_var: gates.INSTR_U(np.pi / 2, phi, lambda_var),
                "u1": lambda lambda_var: gates.INSTR_U(0, 0, lambda_var),
                "cx": instr.INSTR_CNOT,
                "id": gates.INSTR_IDENTITY,
                "u0": gates.INSTR_U(
                    0, 0, 0
                ),  # recently added COMMENT out if causes issues
                "u": gates.INSTR_U,  # alias of U and u3
                "p": lambda lambda_var: gates.INSTR_U(0, 0, lambda_var),  # alias of u1
                "x": instr.INSTR_X,
                "y": instr.INSTR_Y,
                "z": instr.INSTR_Z,
                "h": instr.INSTR_H,
                "s": instr.INSTR_S,
                "sdg": gates.INSTR_S_DAGGER,
                "t": instr.INSTR_T,
                "tdg": gates.INSTR_T_DAGGER,
                "rx": lambda theta: gates.INSTR_U(theta, -np.pi / 2, np.pi / 2),
                "ry": lambda theta: gates.INSTR_U(theta, 0, 0),
                "rz": gates.instrNop_RZ,
            }
            # =============================================================================
            #             standard_lib = {"u3" : gates.INSTR_U, #alias of U native gate
            #                              "u2" : lambda phi, lambda_var : gates.INSTR_U(np.pi/2, phi, lambda_var),
            #                              "u1" : lambda lambda_var : gates.INSTR_U(0, 0, lambda_var),
            #                              "cx" : instr.INSTR_CNOT,
            #                              "id" : gates.INSTR_IDENTITY,
            #                              "u0" : gates.INSTR_U(0, 0, 0), #recently added COMMENT out if causes issues
            #                              "u" : gates.INSTR_U, #alias of U and u3
            #                              "p" : lambda lambda_var : gates.INSTR_U(0, 0, lambda_var), #alias of u1
            #                              "x" : instr.INSTR_X,
            #                              "y" : instr.INSTR_Y,
            #                              "z" : instr.INSTR_Z,
            #                              "h" : instr.INSTR_H,
            #                              "s" : instr.INSTR_S,
            #                              "sdg" : gates.INSTR_S_DAGGER,
            #                              "t" : instr.INSTR_T,
            #                              "tdg" : gates.INSTR_T_DAGGER,
            #                              "rx" : lambda theta : gates.INSTR_U(theta, -np.pi/2, np.pi/2),
            #                              "ry" : lambda theta : gates.INSTR_U(theta, 0, 0),
            #                              "rz" : gates.instrNop_RZ,
            #                              "cz" : instr.INSTR_CZ,
            #                              "cy" : gates.INSTR_CY,
            #                              "ch" : gates.INSTR_CH,
            #                              #"ccx" : instr.INSTR_CCX,
            #                              "crz" : lambda angle : gates.instrNop_RZ(angle, controlled=True),
            #                              "cu1" : lambda lambda_var : gates.INSTR_U(0, 0, lambda_var, controlled=True),
            #                              "cp" : lambda lambda_var : gates.INSTR_U(0, 0, lambda_var, controlled=True), #alias of cu1
            #                              "cu3" : lambda theta, phi, lambda_var : gates.INSTR_U(theta, phi, lambda_var, controlled=True)}
            # =============================================================================
            self.dqc_circuit.native_gates.update(standard_lib_aliases4_native_gates)
            standard_lib_macros = {
                "cz": macros.cz_macro,
                "cy": macros.cy_macro,
                "swap": macros.swap_macro,
                "ch": macros.ch_macro,
                "ccx": macros.ccx_macro,
                "cswap": macros.cswap_macro,
                "crx": macros.crx_macro,
                "cry": macros.cry_macro,
                "crz": macros.crz_macro,
                "cu1": macros.cu1_macro,
                "cp": macros.cp_macro,
                "cu3": macros.cu3_macro,
                "csx": macros.csx_macro,
                "cu": macros.cu_macro,
                "rxx": macros.rxx_macro,
                "rzz": macros.rzz_macro,
                "rccx": macros.rccx_macro,
                "rc3x": macros.rc3x_macro,
                "c3x": macros.c3x_macro,
                "c3sqrtx": macros.c3sqrtx_macro,
                "c4x_macro": macros.c4x_macro,
            }
            self.dqc_circuit.gate_macros.update(standard_lib_macros)
            return self.dqc_circuit
        else:
            raise NotImplementedError(
                "The inclusion of arbitrary files within"
                " the QASM code is not yet supported"
            )


class QasmTwoUniversalSet:
    """The native gates for openQASM 2.0.

    See https://arxiv.org/abs/1707.03429 )

    Attributes
    ----------
    gates : dict
        Lookup table relating strings that reference gates to the relevant
        :class:`~netsquid.components.instructions.Instruction` or function
        that outputs an
        :class:`~netsquid.components.instructions.Instruction`,
        :class:`~netsquid.qubits.operators.Operator` pair.

    """

    gates = {"U": gates.INSTR_U, "CX": instr.INSTR_CNOT}


class Ast2DqcCircuitTranslator:
    """
    Parameters
    ----------
    ast : nested dict
        An abstract syntax tree created from a openQASM 2.0 file using
        :func: `~dqc_simulator.software.qasm2ast.qasm2ast`.
    native_gates : dicts
        A lookup table defining the gates
        that should be treated as native gates for the hardware. The lookup
        table relates strings to the relevant
        :class:`~netsquid.components.instructions.Instruction` or function
        that outputs an
        :class:`~netsquid.components.instructions.Instruction`,
        :class:`~netsquid.qubits.operators.Operator` pair.

    """

    def __init__(self, ast, native_gates=None):
        self.ast = ast
        if native_gates is None:
            self.native_gates = QasmTwoUniversalSet.gates
        else:
            self.native_gates = native_gates

    # =============================================================================
    #     def _interpret_fixed_subgate_params(self, subgate, gate_params,
    #                                         param_interpreter):
    #         interpretted_params = dict()
    #         if gate_params != None:
    #             if subgate['op_param_list'] != None:
    #                 for ii, subgate_param in enumerate(subgate['op_param_list']):
    #                     if subgate_param not in gate_params:
    #                         interpretted_param = param_interpreter(
    #                                                         subgate_param)
    #                         interpretted_params[ii] = interpretted_param
    #                         del subgate['op_param_list'][ii] #so that only non numerical
    #                                                          #arguments are left
    #         elif gate_params == None:
    #             if subgate['op_param_list'] != None:
    #                 for ii, subgate_param in enumerate(subgate['op_param_list']):
    #                     interpretted_param = param_interpreter(
    #                                                     subgate_param)
    #                     interpretted_params[ii] = interpretted_param
    #                     del subgate['op_param_list'][ii] #so that only non numerical
    #                                                      #arguments are left
    #         return interpretted_params
    # =============================================================================

    # =============================================================================
    #     def _ast_g_sect2macros(self, dqc_circuit):
    #         param_interpreter = ExpQasmInterpreter().interpret
    #         for gate_def in self.ast['g_sect']:
    #             gate_name = QasmParsingElement.idQasm.parse_string(
    #                             gate_def['gate_name'])[0]
    #             gate_params = gate_def['gate_param_list']
    #             gate_args = gate_def['gate_reg_list']
    #             if gate_name not in dqc_circuit.native_gates:
    #                 #creating macro:
    #                 subgates = []
    #                 for subgate in gate_def['gate_ops_list']:
    #                     subgate_name = subgate['op']
    #                     if subgate_name in dqc_circuit.gate_macros:
    #                         subgate_params = subgate['op_param_list']
    #                         subgate_args = subgate['op_reg_list']
    #                         if subgate_params is not None:
    #                             expanded_subgate = dqc_circuit.gate_macros[
    #                                 subgate_name](*subgate_params, *subgate_args)
    #                                 #should be list
    #                         else:
    #                             expanded_subgate = dqc_circuit.gate_macros[
    #                                 subgate_name](*subgate_args) #should be list
    #                         subgates = subgates + expanded_subgate #concatenating
    #                                                                #lists
    #                     elif subgate_name in dqc_circuit.native_gates:
    #                         subgates.append(subgate) #appending dict
    #                     else:
    #                         raise ValueError(f'{subgate} not in '
    #                                          'dqc_circuit.gate_macros or '
    #                                          'dqc_circuit.native_gates and so is'
    #                                          'not defined.')
    #                 if gate_params is not None:
    #                     dqc_circuit.gate_macros[gate_name] = lambda gate_params, gate_args :
    # =============================================================================

    def _bake_in_params(
        self,
        subgate_name,
        all_subgate_params,
        interpreted_params,
        uninterpreted_params,
        gate_def_dict,
    ):
        # TO DO: add back in uninterpreted_params argument, then for each
        # uninterpreted arg replace only part of it with the alias_param
        # as a numerical string and then interpret the whole thing inside
        # the alias_func. You can USE
        # QasmParsingElement.idQasm.search_string to find all instances of
        # words with no mathematical relevance. Note that for each element in
        # the resulting list the first sub-element of that element is the
        # desired string.
        def alias_func(*alias_params):
            ii = 0
            jj = 0
            kk = 0
            numerical_subgate_params = []
            for subgate_param in all_subgate_params:
                if subgate_param in interpreted_params:
                    numerical_subgate_params.append(interpreted_params[ii])
                    ii = ii + 1
                else:
                    # this block replaces an uninterpreted parameter with
                    # a string of its numerical value and then interprets
                    # the new string as maths. This allows expressions like
                    #'lambda/2', for which part of the string is already
                    # numerical and part of the string will be replaced by
                    # a number when the alias_func is called, to be handled
                    uninterpreted_param = str(uninterpreted_params[jj])
                    # str used above to make copy rather than changing
                    # uninterpreted_params[jj]
                    unknown_pes = QasmParsingElement.idQasm.search_string(
                        uninterpreted_param
                    )
                    num_unknown_strings = len(unknown_pes)
                    partially_interpreted_param = uninterpreted_param
                    for unknown_pe in unknown_pes:
                        # getting string from parsing element
                        unknown_string = unknown_pe[0]
                        partially_interpreted_param = (
                            partially_interpreted_param.replace(
                                unknown_string, str(alias_params[kk])
                            )
                        )
                    interpreted_param = ExpQasmInterpreter().interpret(
                        partially_interpreted_param
                    )
                    numerical_subgate_params.append(interpreted_param)
                    jj = jj + 1
                    kk = kk + num_unknown_strings
            return gate_def_dict[subgate_name](*numerical_subgate_params)

        return alias_func

    def _make_macro_func(self, parent_def_params, parent_def_args, macro_subgates):
        """
        Making a macro function which takes in parameters and arguments for
        a parent gate (eg, ccx) and outputs the subgates that comprise that
        gate with the appropriate parameters and arguments. We are essentially
        taking in a list of subgates with arguments and params, and linking
        those arguments and params to the input to a function to be called
        later.

        Parameters
        ----------
        parent_def_params : list of str
            The gate_param_list at macro definition (not call) time.
        parent_def_args : list of str
            The gate_reg_list at macro definition (not call) time.
        macro_subgates : list of dict
            The subgate specs in the macro. Each subgate has keys
            ['name', 'params', 'args']
        Returns
        -------


        """

        def macro_func(called_params, called_args):
            """
            A gate macro.

            Parameters
            ----------
            called_params : list of str or None
                Params which will be called by the user of the macro (eg,
                called in AstGate).
            called_args : list of str
                Args which will be called by the user of the macro (eg,
                called in AstGate)..

            Returns
            -------
            macro_subgates : list of dict
                List of subgate specifications for this macro, each with the
                appropriately chosen selection of params and args from the
                overall params and args in the parent gate definition.

            """
            subgates = list(macro_subgates)  # creating local version of
            # macro subgates to avoid accidental
            # changes to external list
            param_lookup = None
            if called_params is not None:
                param_key_value_pairs = zip(parent_def_params, called_params)
                param_lookup = {key: value for (key, value) in param_key_value_pairs}
            arg_key_value_pairs = zip(parent_def_args, called_args)
            arg_lookup = {key: value for (key, value) in arg_key_value_pairs}
            # =============================================================================
            #             print(f' arg lookup is {arg_lookup}')
            # =============================================================================
            for subgate in subgates:
                # =============================================================================
                #                 print(f'subgate is {subgate}')
                # =============================================================================
                if subgate["params"] is not None:
                    for ii, param in enumerate(subgate["params"]):
                        if param_lookup is not None:
                            uninterpreted_param = str(param)
                            # isolating variable names from rest of maths expr
                            variable_parsing_elements = (
                                QasmParsingElement.idQasm.search_string(
                                    uninterpreted_param
                                )
                            )
                            for variable_pe in variable_parsing_elements:
                                variable_name = variable_pe[0]
                                if variable_name in param_lookup:
                                    subgate["params"][ii] = subgate["params"][
                                        ii
                                    ].replace(
                                        variable_name, str(param_lookup[variable_name])
                                    )
                for ii, arg in enumerate(subgate["args"]):
                    if arg in arg_lookup:
                        subgate["args"][ii] = arg_lookup[arg]
            return subgates

        return macro_func

    def _interpret_ast_g_sect(self, dqc_circuit):
        """
        Filtering ast['g_sect'] to see if each gate is a macro or a native gate
        and for macros, expanding them to be in terms of native gates only.
        """

        param_interpreter = ExpQasmInterpreter().interpret
        for gate_def in self.ast["g_sect"]:
            # isolating the gate name from what may be a gate name and
            # params:
            gate_name = QasmParsingElement.idQasm.parse_string(gate_def["gate_name"])[0]
            gate_params = gate_def["gate_param_list"]
            gate_args = gate_def["gate_reg_list"]
            subgates = gate_def["gate_ops_list"]
            if len(subgates) == 1:  # if alias of native gate (potentially with
                # some parameters fixed):
                subgate = subgates[0]
                subgate_name = subgate["op"]
                subgate_params = subgate["op_param_list"]
                interpreted_params = []
                uninterpreted_params = []
                all_subgate_params = []
                if subgate_params is None:
                    dqc_circuit.native_gates[gate_name] = dqc_circuit.native_gates[
                        subgate_name
                    ]
                else:
                    for subgate_param in subgate_params:
                        try:  # if param is interpretable maths expression or number:
                            interpreted_param = param_interpreter(subgate_param)
                            interpreted_params.append(interpreted_param)
                            all_subgate_params.append(interpreted_param)
                        except (
                            Exception
                        ):  # elif just string indicating some parameter name:
                            uninterpreted_params.append(subgate_param)
                            all_subgate_params.append(subgate_param)
                    # =============================================================================
                    #                     #Taking snapshot of relevant info for alias_func definition
                    #                     #to avoid recursive entry to namespace
                    #                     gate_subgate_info = {}
                    #                     gate_subgate_info[gate_name + subgate_name] = {
                    #                         'subgate_params' : list(subgate_params),
                    #                         'uninterpretable_params' : list(uninterpretable_params)}
                    # =============================================================================
                    dqc_circuit.native_gates[gate_name] = self._bake_in_params(
                        subgate_name,
                        all_subgate_params,
                        interpreted_params,
                        uninterpreted_params,
                        dqc_circuit.native_gates,
                    )
                # Note: it does not matter that the name 'alias_func'
                # is re-used, the correct definition is retained
            elif len(subgates) > 1:  # if macro of previously defined gates:
                macro_subgates = []

                for subgate in subgates:
                    subgate_name = subgate["op"]
                    subgate_params = subgate["op_param_list"]
                    subgate_args = subgate["op_reg_list"]

                    if subgate_name in dqc_circuit.native_gates:
                        if subgate_params is None:
                            macro_subgate = {
                                "name": subgate_name,
                                "params": subgate_params,
                                "args": subgate_args,
                            }
                        else:
                            interpreted_params = []
                            uninterpreted_params = []
                            all_subgate_params = []
                            for subgate_param in subgate_params:
                                try:  # if param is interpretable maths expression or number:
                                    interpreted_param = param_interpreter(subgate_param)
                                    interpreted_params.append(interpreted_param)
                                    all_subgate_params.append(interpreted_param)
                                except (
                                    Exception
                                ):  # elif just string indicating some parameter name:
                                    uninterpreted_params.append(subgate_param)
                                    all_subgate_params.append(subgate_param)
                            # making single-use function which is subgate but
                            # with some params baked into it:
                            baked_func_key = gate_name + "_" + subgate_name
                            # creating single-use function and adding to native
                            # gates:
                            dqc_circuit.native_gates[baked_func_key] = (
                                self._bake_in_params(
                                    subgate_name,
                                    all_subgate_params,
                                    interpreted_params,
                                    uninterpreted_params,
                                    dqc_circuit.native_gates,
                                )
                            )
                            macro_subgate = {
                                "name": baked_func_key,
                                "params": subgate_params,
                                "args": subgate_args,
                            }
                        # =============================================================================
                        #                             print(f'native with params, macro subgate is {macro_subgate}')
                        #                             print(' ')
                        # =============================================================================
                        macro_subgates.append(macro_subgate)
                    elif subgate_name in dqc_circuit.gate_macros:
                        # TO DO: fix issue here. cx gates produced here seem to
                        # have repeated args
                        list_of_subgates = dqc_circuit.gate_macros[subgate_name](
                            subgate_params, subgate_args
                        )
                        # =============================================================================
                        #                         print(f'in gate macros for {subgate_name} subgates are: {list_of_subgates}')
                        #                         print(' ')
                        # =============================================================================
                        macro_subgates = macro_subgates + list_of_subgates
                # =============================================================================
                #                 print(f"macro_subgates : {macro_subgates}")
                #                 print(' ')
                # =============================================================================
                dqc_circuit.gate_macros[gate_name] = self._make_macro_func(
                    gate_params, gate_args, macro_subgates
                )

    # =============================================================================
    #                 interpreted_params = []
    #                 uninterpretable_params = []
    #                 for ii, subgate_param in enumerate(subgate_params):
    #                     try: #if param is interpretable maths expression or number:
    #                         interpreted_param = param_interpreter(subgate_param)
    #                         interpreted_params.append(interpreted_param)
    #                         subgate_params[ii] = interpreted_param
    #                     except: #elif just string indicating some parameter name:
    #                         uninterpretable_params.append(subgate_param)
    # =============================================================================
    # =============================================================================
    #                 dqc_circuit.gate_macros[gate_name] =
    # =============================================================================

    # should macro be list of functions with any constant
    # values already incorporated?

    # =============================================================================
    #     def _interpret_ast_g_sect(self, dqc_circuit):
    #         """
    #         filtering ast['g_sect'] to see if each gate is a macro or a native gate
    #         and for macros, expanding them to be in terms of native gates only.
    #         """
    #         param_interpreter = ExpQasmInterpreter().interpret
    #         for gate_def in self.ast['g_sect']:
    #             #isolating the gate name from what may be a gate name and
    #             #params:
    #             gate_name = QasmParsingElement.idQasm.parse_string(
    #                             gate_def['gate_name'])[0]
    #             gate_params = gate_def['gate_param_list']
    #             gate_args = gate_def['gate_reg_list']
    #             if (gate_name not in dqc_circuit.native_gates and
    #                 gate_name not in dqc_circuit.gate_macros):
    #             #if gate is non-native AND hasn't already been added to the dict
    #             #of macros:
    #                 subgates = gate_def['gate_ops_list']
    #                 macro_subgates = []
    # # =============================================================================
    # #                 if len(subgates) == 1 and subgates[0]['op_param_list'] is not None:
    # #                 #if subgate is just alias of native gate or native gate with
    # #                 #some of the parameters fixed:
    # #                     subgate = subgates[0]
    # #                     subgate_name = subgate['op']
    # #                     interpretted_params = dict()
    # #                     if gate_params != None:
    # #                         for ii, subgate_param in enumerate(subgate['op_param_list']):
    # #                             if subgate_param not in gate_params:
    # #                                 interpretted_param = param_interpreter(
    # #                                                                 subgate_param)
    # #                                 interpretted_params[ii] = interpretted_param
    # #
    # #                         native_gate_func = dqc_circuit.native_gates[subgate_name]
    # #                         wrapper = create_wrapper_with_some_args_fixed(
    # #                                                             native_gate_func,
    # #                                                             interpretted_params)
    # #                         dqc_circuit.native_gates[subgate_name] = wrapper
    # #                     elif gate_params == None:
    # #                         for subgate_param in subgate['op_param_list']:
    # #                             subgate_param = param_interpreter(subgate_param)
    # #                         dqc_circuit.native_gates[subgate_name] = dqc_circuit.native_gates[subgate_name]
    # #
    # # =============================================================================
    # # =============================================================================
    # #                 elif len(subgates) > 1:
    # # =============================================================================
    #                 for subgate in subgates:
    #                     subgate_name = subgate['op']
    #
    #                     if subgate_name in dqc_circuit.native_gates:
    #                         interpret = self._interpret_fixed_subgate_params
    #                         interpretted_params = interpret(subgate, gate_params,
    #                                                         param_interpreter)
    #                         #note the above removes the interpretted params from
    #                         #the subgate's 'op_param_list'
    #                         native_gate_func = dqc_circuit.native_gates[
    #                                                                 subgate_name]
    #                         subgate_func = create_wrapper_with_some_args_fixed(
    #                                                             native_gate_func,
    #                                                             interpretted_params)
    #                         subgate_info = {'subgate_func' : subgate_func,
    #                                         'op_param_list' : subgate['op_param_list'],
    #                                         'op_reg_list' : subgate['op_reg_list']}
    #                         macro_subgates.append(subgate_info)
    #                     elif subgate_name in dqc_circuit.gate_macros:
    #                         interpret = self._interpret_fixed_subgate_params
    #                         interpretted_params = interpret(subgate, gate_params,
    #                                                         param_interpreter)
    #                         subgate_params = subgate['op_param_list']
    #                         #copying the macros for this subgate:
    #                         #TO DO: instead make each entry in dqc_circuit.gate_macros
    #                         #a dictionary containing, the gate_params, gate_args
    #                         #and a function, which should take in the gate_params
    #                         #and gate_args (once they have been swapped with their
    #                         #numerical values inside AstGate) and then return a
    #                         #list of gate_specs?
    #                         general_gate_func = dqc_circuit.gate_macros[subgate_name]
    #                         subgate_func = create_wrapper_with_some_args_fixed(
    #                                                             general_gate_func,
    #                                                             interpretted_params)
    #                         subgate_info = {'subgate_func' : subgate_func,
    #                                         'op_param_list' : subgate['op_param_list'],
    #                                         'op_reg_list' : subgate['op_reg_list']}
    #                         macro_subgates.append(subgate_info)
    #
    # # =============================================================================
    # #                             #To avoid having subgate_params overwritten when
    # #                             #a gate macro is simply a call of a native gate with
    # #                             #some parameters fixed:
    # #                             if len(subgate_macros) == 1:
    # #                                 print(f"for gate {gate_name} with subgate {subgate_name}")
    # #                                 print("subgate_macros[0]['op_param_list'] is "
    # #                                       f"{subgate_macros[0]['op_param_list']}")
    # #                                 print(f"this is replaced with {subgate_params}")
    # #     # =============================================================================
    # #     #                             subgate_macros[0]['op_param_list'] = subgate_params
    # #     # =============================================================================
    # #                                 #BUG: this is replacing too much and changing the
    # #                                 #number of params. The issue is that the params are
    # #                                 #right for inserting into the macro but not for the
    # #                                 #underlying native gatea represented by the macro.
    # # =============================================================================
    #
    #
    #                     else:
    #                         raise ValueError(f"Subgate {subgate_name} has not "
    #                                          "previously been defined as a native "
    #                                          "gate or circuit identity (macro)")
    #                 dqc_circuit.gate_macros[gate_name] = {'params' : gate_params,
    #                                                       'args' : gate_args,
    #                                                       'subgates': macro_subgates}
    #                 #should macro_subgates be list of functions with any constant
    #                 #values already incorporated
    # =============================================================================

    def ast2dqc_circuit(self):
        """
        Transferring info from an AST to a form the simulator can work with.

        Converts an abstract syntax tree created using
        :func: `~dqc_simulator.software.qasm2ast.qasm2ast` into a
        :class:`~dqc_simulator.software.dqc_circuit.DqcCircuit` object.

        Returns
        -------
        :class:`~dqc_simulator.software.dqc_circuit.DqcCircuit`
        """

        # implementing functions from <exp> in openQASM 2.0's grammar:
        # TO DO: ascertaining if any of the items in the following dict are
        # superfluous because they will not appear in ast['c_sect'].
        qregs = dict()
        cregs = dict()
        ops4circuit = []
        dqc_circuit = DqcCircuit(qregs, cregs, self.native_gates, ops4circuit)
        # =============================================================================
        #         self._interpret_ast_g_sect(dqc_circuit) #comment out to run code as not
        #                                                 #yet working
        # =============================================================================
        ast_types2sim_types_lookup = {
            ASTType.UNKNOWN: AstUnknown,
            ASTType.COMMENT: AstComment,
            ASTType.QREG: AstQreg,
            ASTType.CREG: AstCreg,
            ASTType.MEASURE: AstMeasure,
            ASTType.BARRIER: AstBarrier,
            ASTType.OP: AstGate,
            ASTType.CTL: AstCtl,
            ASTType.CTL_2: AstCtl2,
            ASTType.BLANK: AstBlank,
            ASTType.DECLARATION_QASM_2_0: AstDeclaration_Qasm_2_0,
            ASTType.INCLUDE: AstInclude,
        }

        for ast_c_sect_element in self.ast["c_sect"]:
            ast2sim_readable_converter = ast_types2sim_types_lookup[
                ast_c_sect_element["type"]
            ](ast_c_sect_element, dqc_circuit)
            # updating dqc_circuit object with new information
            dqc_circuit = ast2sim_readable_converter.make_sim_readable()
        return dqc_circuit


# =============================================================================
# def qasm2sim_readable(filepath):
#     registers = dict()
#     qasm2python_gate_dict = {"U" : gates.INSTR_U,
#                              "CX" : instr.INSTR_CNOT}
#     #implementing functions from <exp> in openQASM 2.0's grammar
#     exp_dict = {"pi" : np.pi, "exp" : np.exp}
#
#
#
#     def _handle_two_qubit_gate(gate_name, params, gate_args, gate_list,
#                                   qasm2python_gate_dict, registers):
#         arg1 = gate_args[0]
#         arg2 = gate_args[1]
#         if '[' in arg1 and '[' in arg2:
#             arg1_reg, arg1_qindex = _split_arg_name_from_index(arg1)
#             arg2_reg, arg2_qindex = _split_arg_name_from_index(arg2)
#             gate_list.append(
#                     (_get_gate_cmd(gate_name, params), arg1_qindex,
#                      arg1_reg, arg2_qindex, arg2_reg))
#         elif '[' not in arg1 and '[' not in arg2:
#             if registers[arg1] != registers[arg2]:
#                 raise ValueError("registers {arg1} and {arg2} have different "
#                                  f" sizes. Their sizes are {registers[arg1]} "
#                                  f"and {registers[arg2]}, respectively.")
#             else:
#                 for ii in range(registers[arg2]):
#                     gate_list.append(
#                         (_get_gate_cmd(gate_name, params), ii, arg1, ii, arg2))
#         elif '[' in arg1 and '[' not in arg2:
#             arg1_reg, arg1_qindex = _split_arg_name_from_index(arg1)
#             for ii in range(registers[arg2]):
#                 gate_list.append(
#                     (_get_gate_cmd(gate_name, params), arg1_qindex, arg1_reg,
#                      ii, arg2))
#         elif '[' not in arg1 and '[' in arg2:
#             arg2_reg, arg2_qindex = _split_arg_name_from_index(arg2)
#             for ii in range(registers[arg1]):
#                 gate_list.append(
#                     (_get_gate_cmd(gate_name, params), ii, arg1, arg2_qindex,
#                      arg2_reg))
#
#
#     with open(filepath, 'r') as file:
#
#         for line in file:
#             word_list = line.split() #TO DO: replace with something using pyparsers parse_string
#             first_word = word_list[0]
#             params = None
#             gate_list = []
#             if '(' in first_word:
#                 split_first_word = first_word.split('(', 1)
#                 first_word = split_first_word[0]
#                 params = split_first_word[1].replace(')', '').split(',') #params should now be list of parameters
#             if first_word == 'OPENQASM':
#                 pass
#             elif first_word == '//':
#                 pass
#             elif first_word == 'include':
#                 #standard header is so ubiquitous that it makes sense to handle
#                 #it as its own case
#                 if word_list[1] == "qelib1.inc":
#                     standard_lib = {"u3" : gates.INSTR_U,
#                                      "u2" : lambda phi, lambda_var : gates.INSTR_U(np.pi/2, phi, lambda_var),
#                                      "u1" : lambda lambda_var : gates.INSTR_U(0, 0, lambda_var),
#                                      "cx" : instr.INSTR_CNOT,
#                                      "id" : gates.INSTR_IDENTITY,
#                                      "x" : instr.INSTR_X,
#                                      "y" : instr.INSTR_Y,
#                                      "z" : instr.INSTR_Z,
#                                      "h" : instr.INSTR_H,
#                                      "s" : instr.INSTR_S,
#                                      "sdg" : gates.INSTR_S_DAGGER,
#                                      "t" : instr.INSTR_T,
#                                      "tdg" : gates.INSTR_T_DAGGER,
#                                      "rx" : lambda theta : gates.INSTR_U(theta, -np.pi/2, np.pi/2),
#                                      "ry" : lambda theta : gates.INSTR_U(theta, 0, 0),
#                                      "rz" : gates.INSTR_RZ,
#                                      "cz" : instr.INSTR_CZ,
#                                      "cy" : gates.INSTR_CY,
#                                      "ch" : gates.INSTR_CH,
#                                      "ccx" : instr.INSTR_CCX,
#                                      "crz" : lambda angle : gates.INSTR_RZ(angle, controlled=True),
#                                      "cu1" : lambda lambda_var : gates.INSTR_U(0, 0, lambda_var, controlled=True),
#                                      "cu3" : lambda theta, phi, lambda_var : gates.INSTR_U(theta, phi, lambda_var, controlled=True)}
#                     qasm2python_gate_dict.update(standard_lib)
#                 else:
#                     pass #TO DO: add ability to include files
#             elif first_word == 'qreg' or first_word == 'creg':
#                 reg_def = word_list[1].split('[')
#                 reg_name = reg_def[0]
#                 reg_size = reg_def[1].split(']')[0]
#                 registers[reg_name] = int(reg_size)
#             elif first_word == 'barrier':
#                 pass #TO DO: replace pass with relevant code
#             elif first_word == 'measure':
#                 pass #TO DO: replace pass with relevant code
#             elif first_word in qasm2python_gate_dict: #first_word is quantum gate
#                 gate_args = word_list[1::]
#                 if len(gate_args) == 1:
#                     _handle_single_qubit_gate(first_word, params, gate_args,
#                                               gate_list, qasm2python_gate_dict,
#                                               registers)
#                 elif len(gate_args) == 2:
#                     _handle_two_qubit_gate(first_word, params, gate_args,
#                                            gate_list, qasm2python_gate_dict,
#                                            registers)
#                 elif first_word == "ccx":
#                     pass #TO DO: NEED TO IMPLEMENT something here
#
#     return gate_list
# =============================================================================


# do not need to close file if you use 'with open()' rather than just
#'open()'


# TO DO: GENERALISE code by fixing WARNING comments

# TO DO:FIGURE out how to handle case where different qregs are on the same node and so
# the qreg name does not correspond to the node name, like you have been assuming.
# This can be done at the end (potentially in the compiler) but if not in a
# post-processing step where the number of different qregs is evaluated and they
# are assigned to nodes. In cases where there is only 1 qreg, you can automatically
# say the circuit is monolithic.

# TO DO: add ability to handle include statements


# Cases not handled:
#   1) Where bracket is



# -*- coding: utf-8 -*-
# =============================================================================
# Created on Wed Mar 20 09:52:29 2024
#
# @author: kenny
# =============================================================================
"""Macros for useful quantum circuits."""

from netsquid.components import instructions as instr

from dqc_simulator.qlib.gates import INSTR_T_DAGGER


def get_ghz_gate_tuples(num_qubits):
    """
    Return the gate tuples for a generalised GHZ gate implemented on a
    monolithic quantum processor.

    Parameters
    ----------
    num_qubits : int
        The number of qubits to include in the generalised GHZ state.

    Returns
    -------
    gate_tuples : list of tuple
        The gate_tuples needed to simulate a quantum circuit on a monolithic
        quantum processor.
    """
    gate_tuples = [
        (instr.INSTR_INIT, [ii for ii in range(num_qubits)], "mono_qc"),
        (instr.INSTR_H, 0, "mono_qc"),
        *[
            (instr.INSTR_CNOT, ii, "mono_qc", ii + 1, "mono_qc")
            for ii in range(num_qubits - 1)
        ],
    ]
    return gate_tuples


def get_gate_tuples_to_create_cluster_state(topology):
    """
    Creates list of cluster state gates for implementation on a single-QPU
    quantum computer.

    Parameters
    ----------
    topology : list of tuples of ints
        Each tuple should contain the two indices, one for each qubit connected
        by an edge (CZ gate)

    Returns
    -------
    list of tuples
        The gate tuples needed to generate a cluster state with the specified
        topology.
    """
    qubits_to_initialise = []
    cz_gates = []
    for edge in topology:
        qubit_index0 = edge[0]
        qubit_index1 = edge[1]
        gate_tuple = (instr.INSTR_CZ, qubit_index0, "mono_qc", qubit_index1, "mono_qc")
        cz_gates.append(gate_tuple)
        if qubit_index0 not in qubits_to_initialise:
            qubits_to_initialise.append(qubit_index0)
        if qubit_index1 not in qubits_to_initialise:
            qubits_to_initialise.append(qubit_index1)
    qubits_to_initialise.sort()
    init_cmd = [(instr.INSTR_INIT, qubits_to_initialise, "mono_qc")]
    hadamards = [
        (instr.INSTR_H, qubit_index, "mono_qc") for qubit_index in qubits_to_initialise
    ]
    return init_cmd + hadamards + cz_gates


def produce_partitioned_circuit_from_fig4_in_autocomm_paper():
    """
    Macro for a specific circuit (see references)

    Returns
    -------
    partitioned_gates : list of tuples
        The gates needed to describe the monolithic quantum circuit.

    Notes
    -----
    This circuit can be used to test implementations of AutoComm [1]_ and
    other compilers (which can be compared to AutoComm)

    References
    ----------
    See figure 4 of [1]_ for an diagram of the circuit in question.

    .. [1] A. Wu, H. Zhang, G. Li, A. Shabani, Y. Xie, and Y. Ding, Autocomm:
           A framework for enabling efficient communication in
           distributed quantum programs, in 2022 55th IEEE/ACM International
           Symposium on Microarchitecture (MICRO) (2022) pp. 1027–1041.
    """
    partitioned_gates = [
        (instr.INSTR_T, 2, "node_1"),
        (instr.INSTR_CNOT, 2, "node_0", 2, "node_1"),
        (INSTR_T_DAGGER, 2, "node_1"),
        (instr.INSTR_T, 3, "node_1"),
        (instr.INSTR_T, 3, "node_2"),
        (instr.INSTR_CNOT, 2, "node_0", 2, "node_1"),
        (instr.INSTR_CNOT, 3, "node_0", 3, "node_1"),
        (instr.INSTR_CNOT, 2, "node_0", 3, "node_2"),
        (INSTR_T_DAGGER, 3, "node_0"),
        (INSTR_T_DAGGER, 3, "node_1"),
        (INSTR_T_DAGGER, 3, "node_2"),
        (instr.INSTR_CNOT, 3, "node_0", 3, "node_1"),
        (instr.INSTR_CNOT, 2, "node_0", 3, "node_2"),
        (instr.INSTR_H, 2, "node_2"),
        (instr.INSTR_CNOT, 2, "node_2", 3, "node_1"),
        (INSTR_T_DAGGER, 3, "node_1"),
        (instr.INSTR_H, 2, "node_2"),
        (instr.INSTR_CNOT, 2, "node_2", 3, "node_1"),
        (instr.INSTR_CNOT, 2, "node_2", 2, "node_1"),
        (instr.INSTR_T, 2, "node_2"),
        (instr.INSTR_CNOT, 3, "node_0", 2, "node_2"),
        (INSTR_T_DAGGER, 3, "node_0"),
        (instr.INSTR_T, 2, "node_2"),
        (instr.INSTR_CNOT, 2, "node_2", 3, "node_0"),
        (instr.INSTR_CNOT, 2, "node_2", 2, "node_0"),
        (INSTR_T_DAGGER, 3, "node_0"),
        (instr.INSTR_T, 3, "node_1"),
        (instr.INSTR_CNOT, 3, "node_0", 2, "node_1"),
        (instr.INSTR_CNOT, 2, "node_2", 3, "node_1"),
        (INSTR_T_DAGGER, 3, "node_0"),
        (instr.INSTR_T, 2, "node_1"),
        (INSTR_T_DAGGER, 3, "node_1"),
        (instr.INSTR_T, 2, "node_2"),
        (instr.INSTR_CNOT, 2, "node_1", 3, "node_0"),
        (instr.INSTR_CNOT, 3, "node_1", 2, "node_2"),
        (instr.INSTR_CNOT, 2, "node_0", 2, "node_1"),
    ]
    return partitioned_gates



# -*- coding: utf-8 -*-
# =============================================================================
# Created on Wed Jan 31 12:26:58 2024
#
# @author: kenny
# =============================================================================
"""
Info on each remote gate scheme.
"""


def get_scheme_info():
    # The resource costs associated with implementing a single remote CX using
    # each of the remote gate schemes. The 'num_single_qubit_gates' assumes that all
    # measurement results are 1 (the worst case scenario).
    return {
        "mono": {
            "num_epr_pairs": 0,
            "num_cnots": 1,
            "num_single_qubit_gates": 0,
            "num_measurements": 0,
            "num_classical_comms": 0,
        },
        "cat": {
            "num_epr_pairs": 1,
            "num_cnots": 2,
            "num_single_qubit_gates": 3,
            "num_measurements": 2,
            "num_classical_comms": 2,
        },
        "1tp": {
            "num_epr_pairs": 1,
            "num_cnots": 2,
            "num_single_qubit_gates": 3,
            "num_measurements": 2,
            "num_classical_comms": 2,
        },
        "2tp": {
            "num_epr_pairs": 2,
            "num_cnots": 3,
            "num_single_qubit_gates": 6,
            "num_measurements": 4,
            "num_classical_comms": 4,
        },
        "tp_safe": {
            "num_epr_pairs": 2,
            "num_cnots": 6,
            "num_single_qubit_gates": 6,
            "num_measurements": 4,
            "num_classical_comms": 4,
        },
    }



# -*- coding: utf-8 -*-
"""
Tools for the generation of a distributed quantum computer network.

These tools are designed to be controlled using software from
:mod: `~dqc_simulator.software.dqc_control`
"""

import warnings

from netsquid.nodes import Node, DirectConnection, Network
from netsquid.components.cchannel import ClassicalChannel
from netsquid.qubits import ketstates as ks
from netsquid.components.models.delaymodels import FibreDelayModel

from dqc_simulator.hardware.connections import (
    create_black_box_central_source_entangling_link,
    create_classical_fibre_link,
)
from dqc_simulator.hardware.quantum_processors import (
    create_qproc_with_analytical_noise_ionQ_aria_durations_N_standard_lib_gates,
)
from dqc_simulator.util.helper import (
    create_wrapper_with_some_args_fixed,
    filter_kwargs4internal_functions,
)


def link_2_qpus(
    network,
    node_a,
    node_b,
    want_classical_2way_link=True,
    want_extra_classical_2way_link=None,  # can make scheduling
    # easier
    want_entangling_link=True,
    create_entangling_link=None,
    # kwargs for classical links and default entangling link
    node_distance=2e-3,
    # kwargs for default entangling link
    state4distribution=None,
    ent_dist_rate=0,
    # kwargs for non-default entangling_link
    **kwargs4create_entangling_link,
):
    """
    Sets up a link between two QPUs.

    Adds a :class:`~netsquid.netsquid.nodes.connections.Connection` between
    two QPUs and sets up input forwarding between ports on the QPUs so that
    qubits are sent to the right places.

    Parameters
    ----------
    network : netsquid.nodes.network.Network
        The entire network.
    node_a :  object created using netsquid.nodes.Node class
    A network node
    node_b : object created using netsquid.nodes.Node class
    A network node
    create_entangling_link : func, optional
        The function to use in order to create an entangling link. By default,
        :func: `create_black_box_central_source_entangling_link` is used.
    node_distance : float, optional
        Distance between adjacent nodes in km. Used by default value of
        `create_entangling_link`. May not be used, if a different
        `create_entangling_link` function is used.
    want_classical_2way_link : bool, optional
        Whether a two-way classical link should be created between all nodes
        (True) or not (False). Default is True.
    want_extra_classical_2way_link : bool, optional
        Whether an extra two-way classical link should be created between all
        nodes (True) or not (False). Default is True.
    want_entangling_link : bool, optional
        Whether a two-way quantum link should be created between all nodes
        (True) or not (False). Nodes can request entangled pairs using this
        link.
    state4distribution : numpy.ndarray
        The entangled state distributed between nodes when
        an EPR pair is requested. Default is |phi^+> Bell state (formalism not
        fixed to ket). `state4distribution` is used by the default
        `create_entangling_link` function and may be ignored if a different
        entangling link is used.
    ent_dist_rate : float, optional
        The rate of entanglement distribution [Hz]. Used by default value of
        `create_entangling_link`. May not be used, if a different
        `create_entangling_link` function is used. `ent_dist_rate` is used by
        the default `create_entangling_link` function and may be used be
        ignored if a different entangling link is used.
    kwargs4create_entangling_link
        The keyword arguments for the chosen `create_entangling_link` function.
    """
    # instantiating default arguments. Need to do in body of function for mutable
    # objects like a numpy array because default values of functions are defined
    # only when the function is defined (not when it's called) and so mutable
    # default values can be accidentally changed by code outwith the function.
    # See https://stackoverflow.com/questions/10676729/why-does-using-arg-none-
    # fix-pythons-mutable-default-argument-issue
    if want_extra_classical_2way_link is None:
        want_extra_classical_2way_link = want_classical_2way_link
    if state4distribution is None:
        state4distribution = ks.b00
    if (
        create_entangling_link is None
        or create_entangling_link == create_black_box_central_source_entangling_link
    ):
        # The default value (create_black_box_central_source_entangling_link) is
        # wrapped, so that in later code, there is no need to state argument,
        # which is not required by many connections. This keeps the variable
        # name create_entangling_link more generic.
        args2fix = {3: state4distribution}  # fixing 4th argument (index=3) with
        # value state4distribution
        create_entangling_link = create_wrapper_with_some_args_fixed(
            create_black_box_central_source_entangling_link,
            args2fix,
            ent_dist_rate=ent_dist_rate,
        )

    if want_classical_2way_link is False and want_entangling_link is False:
        raise ValueError("""At least one of want_classical_2way_link and
                         want_entangling_link must be True otherwise
                         the function call would be redundant""")
    # Set up classical connection between nodes:
    if want_classical_2way_link:
        # TO DO: give the user option to specify what type of link is used (as
        # they can for the entangling link). Eg, the user may wish to use a
        # free-space classical connection instead.
        create_classical_fibre_link(
            network, node_a, node_b, length=node_distance, label="classical"
        )

    if want_extra_classical_2way_link:
        # This is useful if you want to make scheduling easier by separating out
        # tasks. For example, one link for exchanging measurement results and one
        # for doing handshake protocols.
        # TO DO: give the user option to specify what type of link is used (as
        # they can for the entangling link). Eg, the user may wish to use a
        # free-space classical connection instead.
        create_classical_fibre_link(
            network, node_a, node_b, length=node_distance, label="extra_classical"
        )

    if want_entangling_link:
        create_entangling_link(network, node_a, node_b, **kwargs4create_entangling_link)


def create_dqc_network(
    *args4qproc,
    node_list=None,
    quantum_topology=None,
    classical_topology=None,
    want_classical_2way_link=True,
    want_extra_classical_2way_link=None,
    want_entangling_link=True,
    num_qpus=2,
    name="linear network",
    create_entangling_link=None,
    custom_qpu_func=None,
    # parameter for default classical AND entangling connection
    node_distance=2e-3,
    # parameters for default entangling connection
    state4distribution=None,  # ks.b00 defined in function body
    ent_dist_rate=182,  # Hz
    # parameters for default quantum processor
    num_comm_qubits=2,
    p_depolar_error_cnot=0,
    comm_qubit_depolar_rate=0,
    proc_qubit_depolar_rate=0,
    # for any non-default entangling connection or quantum processor
    **kwargs,
):
    """
    Creates a network suitable for distributed quantum computing

    Creates a network with QPUs node_0 to node_nn
    connected by entangling connections and/or classical connections.
    In other words this function sets up the physical components of the
    distributed quantum computer. The
    default topology of the network is a linear topology of the form
    node_1 <--> node_2 <--> node_3 <--> ...
    If not manually specified with the optional node_list argument, all nodes
    will be identical.

    Parameters
    ----------
    *args4qproc :
        The arguments for the custom_qpu_func specified. By default,
        there are none of these.
    node_list :  list of netsquid.components.nodes.Node objects, optional
        List of nodes to be included in network. Overrides num_qpus if
        specified.
    num_qpus : int (>=2), optional
        The number of identical nodes to be included in the network. Each node
        has 5 data qubits and two communication qubits. Must be at least 2 to
        avoid trivial scenario (if you have no connections there is no need)
        for this function, just create a node.
    Is overridden if a node_list is specified.
    node_distance : float, optional
        Distance between adjacent nodes.
    quantum_toplogy : list of tuple pairs of integers, optional
        The quantum network topology. Default is linear. The indices refer to
        the nodes.to be connected should be connected in the tuple.
    classical_topology: list of tuple pairs of integers, optional
            The classical network topology. Default is linear. The indices
            refer to the nodes.to be connected should be connected in the tuple.
    want_classical_2way_link : bool, optional
        Whether a two-way classical link should be created between all nodes
        (True) or not (False). Default is True. Unused if `classical_topology`
        or `quantum_topology` is not None.
    want_extra_classical_2way_link : bool or None, optional
        Whether an extra two-way classical link should be created between all
        nodes (True) or not (False). Default is None. If None, it will take
        the same value as `want_classical_2way_link`. Unlike,
        `want_classical_2way_link` this will be used if `classical_topology` is
        not None (in which case an extra classical link will be created
        everywhere a classical link is created).
    want_entangling_link : bool, optional
        Whether a two-way quantum link should be created between all nodes
        (True) or not (False). Nodes can request entangled pairs using this
        link. Unused if `classical_topology` or `quantum_topology` is not None.
    create_entangling_link : func, optional
        The function to use in order to create an entangling link. By default,
        :func: `create_black_box_central_source_entangling_link` is used.
    custom_qpu_func: function or None, optional
        Creates the quantum processor object to use on each node. It must
        be able to run with no positional arguments. If unspecified then
        a default noiseless processor will be used
    state4distribution : numpy.ndarray
        The entangled state distributed between nodes when
        an EPR pair is requested. Default is |phi^+> Bell state (formalism not
        fixed to ket). This argument is used by the default
        `create_entangling_link` function but may be ignored if another
         function is specified using `create_entangling_link`.
    ent_dist_rate : float, optional
        The rate of entanglement distribution [Hz]. Used by default value of
        `create_entangling_link`. May not be used, if a different
        `create_entangling_link` function is used. `ent_dist_rate` is used by
        the default `create_entangling_link` function and may be used be
        ignored if a different entangling link is used.
    num_comm_qubits : int, optional
        The number of communication qubit positions on the QPU. Default is 2.
        This argument is used by the default custom_qpu_func but may be ignored
        if a different function is specified using the `create_qpu_func`
        parameter and that function does not use this argument.
    p_depolar_error_cnot : float, optional
        The probability of a depolarisation error after each CNOT gate.
        The default value is 0.
        This argument is used by the default custom_qpu_func but may be ignored
        if a different function is specified using the `create_qpu_func`
        parameter and that function does not use this argument.
    comm_qubit_depolar_rate : float, optional
        The depolarisation rate for communication qubits.
        The default value is 0.
        This argument is used by the default custom_qpu_func but may be ignored
        if a different function is specified using the `create_qpu_func`
        parameter and that function does not use this argument.
    proc_qubit_depolar_rate : float
        The depolarisation rate for processing qubits.
        The default value is 0.
        This argument is used by the default custom_qpu_func but may be ignored
        if a different function is specified using the `create_qpu_func`
        parameter and that function does not use this argument.
    name : str, optional
        Name of the network
    **kwargs :
        The name-value arguments for the custom_qpu_func and
        create_entangling_link functions specified.

    Returns
    -------
    network : :class:`~netsquid.nodes.network.Network`
        The simulated hardware for a distributed quantum computer.
    """

    # initialising default arguments
    if want_extra_classical_2way_link is None:
        want_extra_classical_2way_link = want_classical_2way_link
    if state4distribution is None:
        state4distribution = ks.b00
    if custom_qpu_func is None:
        custom_qpu_func = (
            create_qproc_with_analytical_noise_ionQ_aria_durations_N_standard_lib_gates
        )
    if create_entangling_link is None:
        create_entangling_link = create_black_box_central_source_entangling_link
    # instantiating keyword arguments for default custom_qpu_func
    # function, this is necessary because I have explicitly included those
    # keyword arguments in the function definition above. This is done to
    # make the documentation clearer (ie, explicitly make the user aware of
    # the default options)
    kwargs["num_comm_qubits"] = num_comm_qubits
    kwargs["p_depolar_error_cnot"] = p_depolar_error_cnot
    kwargs["comm_qubit_depolar_rate"] = comm_qubit_depolar_rate
    kwargs["proc_qubit_depolar_rate"] = proc_qubit_depolar_rate
    # instantiating keyword arguments for default create_entangling_link
    # function, this is necessary because I have explicitly included those
    # keyword arguments in the function definition above. This is done to
    # make the documentation clearer (ie, explicitly make the user aware of
    # the default options)
    kwargs["state4distribution"] = state4distribution
    kwargs["node_distance"] = node_distance
    kwargs["ent_dist_rate"] = ent_dist_rate

    kwargs4qproc, kwargs4create_entangling_link = filter_kwargs4internal_functions(
        [custom_qpu_func, create_entangling_link], kwargs
    ).values()

    if num_qpus < 2:
        raise ValueError("num_qpus must be at least 2")
    if type(num_qpus) is not int:
        raise TypeError("num_qpus must be an integer")

    def _create_qpus(num_qpus2create=num_qpus, starting_index=0):
        qpu_list = []
        for ii in range(num_qpus2create):
            # need to instantiate a new processor object each time with a
            # new function call. Otherwise, the code is trying to add the
            # same processor object to different nodes, as opposed to a copy
            # of it. Therefore, it makes most sense to define this here within
            # create_dqc_network
            qpu = custom_qpu_func(*args4qproc, **kwargs4qproc)
            qpu.name = f"qproc4node_{ii + starting_index}"
            qpu_node = Node(f"node_{ii + starting_index}", qmemory=qpu)
            qpu_list = qpu_list + [qpu_node]  # appending node to node_list
        return qpu_list

    def _handle_even_num_qpus(network, num_qpus_considered):
        """
        Subroutine for creating linear network with even number of QPUs

        Parameters
        ----------
        network : TYPE
            DESCRIPTION.
        num_qpus_considered : TYPE
            DESCRIPTION.
        """
        for ii in range(num_qpus_considered - 1):
            node_1 = network.get_node(f"node_{ii}")
            node_2 = network.get_node(f"node_{ii + 1}")
            link_2_qpus(
                network,
                node_1,
                node_2,
                want_classical_2way_link=want_classical_2way_link,
                want_extra_classical_2way_link=want_extra_classical_2way_link,
                want_entangling_link=want_entangling_link,
                create_entangling_link=create_entangling_link,
                **kwargs4create_entangling_link,
            )

    network = Network(name)

    if node_list is None:
        node_list = _create_qpus(num_qpus2create=num_qpus)
    # NetSquid natively raises error if you try adding nodes with the same
    # name to a network, so it is safe to assume each node has a unique name in
    # what follows despite the possibility of mixing and matching auto-generated
    # and user-defined names
    network.add_nodes(node_list)
    if quantum_topology is None and classical_topology is None:
        # create network with linear topology
        if (num_qpus % 2) == 0:  # if num_qpus is even
            _handle_even_num_qpus(network, num_qpus)
        else:  # if num_qpus is odd
            # handle first num_qpus-1 nodes
            num_qpus_considered = num_qpus - 1
            _handle_even_num_qpus(network, num_qpus_considered)
            # handle the rest
            node_1 = network.get_node(f"node_{num_qpus - 2}")
            node_2 = network.get_node(f"node_{num_qpus - 1}")  # num_qpus-1 is
            # max index
            link_2_qpus(
                network,
                node_1,
                node_2,
                want_classical_2way_link=want_classical_2way_link,
                want_extra_classical_2way_link=want_entangling_link,
                want_entangling_link=want_entangling_link,
                create_entangling_link=create_entangling_link,
                **kwargs4create_entangling_link,
            )
    else:
        if quantum_topology is not None:
            if (quantum_topology[-1])[1] > (len(network.nodes) - 1):
                # if index is to large for number of nodes in network
                nodes_needed = (quantum_topology[-1])[1] - (len(network.nodes) - 1)
                extra_nodes = _create_qpus(
                    num_qpus2create=nodes_needed, starting_index=num_qpus
                )
                network.add_nodes(extra_nodes)
                warnings.warn(
                    "WARNING: num_qpus was overwritten because there"
                    " were not enough nodes to realise the requested"
                    " topology"
                )

            for ii in range(len(quantum_topology)):
                indices = quantum_topology[ii]
                node_a = network.get_node(f"node_{indices[0]}")
                node_b = network.get_node(f"node_{indices[1]}")
                link_2_qpus(
                    network,
                    node_a,
                    node_b,
                    want_classical_2way_link=False,
                    want_extra_classical_2way_link=False,
                    want_entangling_link=True,
                    create_entangling_link=create_entangling_link,
                    **kwargs4create_entangling_link,
                )
        if classical_topology is not None:  # if classical_topology is not None
            if (classical_topology[-1])[1] > (len(network.nodes) - 1):
                # if index is too large for number of QPUs in network
                nodes_needed = (classical_topology[-1])[1] - (len(network.nodes) - 1)
                extra_nodes = _create_qpus(
                    num_qpus2create=nodes_needed, starting_index=num_qpus
                )
                network.add_nodes(extra_nodes)
                warnings.warn(
                    "WARNING: num_qpus was overwritten because there"
                    " were not enough nodes to realise the requested"
                    " topology"
                )

            for ii in range(len(classical_topology)):
                indices = classical_topology[ii]
                node_a = network.get_node(f"node_{indices[0]}")
                node_b = network.get_node(f"node_{indices[1]}")
                link_2_qpus(
                    network,
                    node_a,
                    node_b,
                    want_classical_2way_link=True,
                    want_extra_classical_2way_link=want_extra_classical_2way_link,
                    want_entangling_link=False,
                    create_entangling_link=create_entangling_link,
                    **kwargs4create_entangling_link,
                )
    return network


class DQC(Network):
    """
    Encapsulates all of the hardware needed for distributed quantum computing.

    Creates `num_qpus` QPU nodes, with quantum memories defined by `qpu_class`,
    connected by quantum and classical Connections according to the classical
    and quantum topologies specified by `classical_topology` and
    `quantum_topology`, respectively.


    Parameters
    ----------
    entangling_connection_class : subclass of :class:`~netsquid.nodes.connections.Connection`
        The subclass to use for creating a connections between QPU nodes. Can
        be used to define a specific connection. Pre-made Connections can
        be found in `dqc_simulator.hardware.connections`.
    num_qpus : int
        The number of quantum processing units in the DQC.
    quantum_topology: list of (int, int)
        The topology of quantum connections between QPU nodes.
    classical_topology : list of (int, int)
        The topology of classical connections between QPU nodes. These are
        optical fibre connections.
    qpu_class : subclass of :class:`~dqc_simulator.hardware.quantum_processors.QPU` or None, optional
        The class to use when creating the QPUs. Allows a specific type of
        QPU to be specied. The default is None. If this is None, `nodes`
        must not be None and the QPUs used will be defined there.
    classical_connection_class : subclass of :class:`~netsquid.nodes.connections.Connection` or None, optional
        The classical connection to use. The default is None. If None, a
        classical fibre connection is used
    want_extra_classical_link : bool, optional
        Whether to create two classical connections of class
        `classical_connection_class` for every connection indicated in
        `classical_topology`. The default is True.
    name : str, optional
        The name of the DQC object. The default is "DQC".
    nodes : list of :class:`~netsquid.nodes.node.Node` instances or None, optional
        Overwritten if qpu_class is specified. If this is None, `nodes` must
        not be None.
    node_separation : float or None, optional
        The distance between QPU nodes in km. Used when instantiating the default
        `classical_connection_class`. Default is None. If None, the value of
        2e-3 (corresponding to 2m) is used.
    **kwargs
        The kwargs needed by `entangling_connection_class`, `qpu_class` and
        `classical_connection_class`. Note that if these classes have
        keyword arguments with the same name, an error will be raised.


    Returns
    -------
    network : :class:`~netsquid.nodes.network.Network`
        The simulated hardware for a distributed quantum computer.
    """

    def __init__(
        self,
        entangling_connection_class,
        num_qpus,
        quantum_topology,
        classical_topology,
        qpu_class=None,
        classical_connection_class=None,
        want_extra_classical_link=True,
        name="DQC",
        nodes=None,
        node_separation=None,
        **kwargs,
    ):
        super().__init__(name, nodes=nodes)
        # Instantiating mutable default values.
        if node_separation is None:
            self.node_separation = 2e-3  # 2m
        else:
            self.node_separation = node_separation

        # Splitting the kwargs by the class that they should be applied to
        class_tuple = (
            entangling_connection_class,
            qpu_class,
            classical_connection_class,
        )
        classes = [value for value in class_tuple if value is not None]
        kwargs4each_class = filter_kwargs4internal_functions(classes, kwargs)
        kwarg_types = {key: kwargs4each_class[key] for key in classes}

        self.qpu_class = qpu_class
        self.num_qpus = num_qpus

        if nodes is None and qpu_class is None:
            raise ValueError("nodes and qpu_class cannot both have value None")
        if qpu_class is not None:
            qpu_nodes = self._create_qpu_nodes(**kwarg_types[qpu_class])
            self.add_nodes(qpu_nodes)
        else:
            qpu_nodes = list(self.nodes.values())

        for pair in quantum_topology:
            entangling_connection = entangling_connection_class(
                **kwarg_types[entangling_connection_class]
            )
            label = "entangling"
            self._add_link(qpu_nodes, pair, entangling_connection, label)
        for pair in classical_topology:
            label = "classical"
            classical_connection = self._get_classical_connection(
                classical_connection_class, kwarg_types
            )
            self._add_link(qpu_nodes, pair, classical_connection, label)
            if want_extra_classical_link:
                label = "extra_classical"
                classical_connection = self._get_classical_connection(
                    classical_connection_class, kwarg_types
                )
                self._add_link(qpu_nodes, pair, classical_connection, label)

    def _create_qpu_nodes(self, **kwargs4qpu):
        qpu_list = []
        for ii in range(self.num_qpus):
            qpu = self.qpu_class(**kwargs4qpu)
            qpu.name = f"qproc4node_{ii}"
            qpu_node = Node(f"node_{ii}", qmemory=qpu)
            qpu_list = qpu_list + [qpu_node]  # appending node to node_list
        return qpu_list

    def _get_classical_connection(self, classical_connection_class, kwarg_types):
        if classical_connection_class is None:
            return DirectConnection(
                "DirectConnection",
                channel_AtoB=ClassicalChannel(
                    "Channel_A2B",
                    length=self.node_separation,
                    models={"delay_model": FibreDelayModel()},
                ),
                channel_BtoA=ClassicalChannel(
                    "Channel_B2A",
                    length=self.node_separation,
                    models={"delay_model": FibreDelayModel()},
                ),
            )
        else:
            return classical_connection_class(**kwarg_types[classical_connection_class])

    def _add_link(self, qpu_nodes, nodes2connect, connection, label):
        """


        Parameters
        ----------
        qpu_nodes : TYPE
            DESCRIPTION.
        qpus2connect : (int, int)
            The indices of the QPU nodes to connect
        connection : TYPE
            DESCRIPTION.
        label : TYPE
            DESCRIPTION.

        Returns
        -------
        None.

        """
        node_a = qpu_nodes[nodes2connect[0]]
        node_b = qpu_nodes[nodes2connect[1]]
        node_a_port_name = node_a.connection_port_name(node_b.name, label=label)
        node_b_port_name = node_b.connection_port_name(node_a.name, label=label)
        self.add_connection(
            node_a,
            node_b,
            connection=connection,
            label=label,
            port_name_node1=node_a_port_name,
            port_name_node2=node_b_port_name,
        )


# CHANGELOG:
# create_processor changed to have only optional arguments.



# -*- coding: utf-8 -*-
# =============================================================================
# Created on Thu Aug 17 12:28:32 2023
#
# @author: kenny
# =============================================================================
"""
Extendable library of useful quantum states and helper functions for creating
states.
"""

import functools as ft

import netsquid as ns
from netsquid.qubits import ketstates as ks
import numpy as np


def get_zket(*args):
    """
    Define arbitrary tensor product of kets in the computational basis.

    Parameters
    ----------
    *args : int
        The bits defining the basis. Eg, for args 1, 0, 1, basis return |101>

    Returns
    -------
    np.array
        Ket of form |ijkl....> for args i, j, k, l....

    """
    ket_lookup = {0: ks.s0, 1: ks.s1}
    return ft.reduce(np.kron, [ket_lookup[arg] for arg in args])


def get_ghz_state_ket(num_qubits):
    """
    Return a ket vector for the n-qubit generalised GHZ state

    Parameters
    ----------
    num_qubits : int
        The number of qubits, n, in the n-qubit generalised GHZ state.

    Returns
    -------
    :class:`~numpy.ndarray`
        The ket vector for an n-qubit genralised GHZ state.

    """
    return (
        1
        / np.sqrt(2)
        * (
            ft.reduce(np.kron, [ks.s0] * num_qubits)
            + ft.reduce(np.kron, [ks.s0] * num_qubits)
        )
    )


def werner_state(F):
    """
    Defines the density matrix for the Werner state.

    Parameters
    ----------
    F : float
        The Werner state fidelity (fidelity of |Phi^+>) within the Werner
        state.

    Returns
    -------
    werner_state : array or array-like
        The density matrix for the Werner state.
    """
    werner_state = (
        F * ns.qubits.ketutil.ket2dm(ks.b00)
        + (1 - F) / 3 * ns.qubits.ketutil.ket2dm(ks.b01)
        + (1 - F) / 3 * ns.qubits.ketutil.ket2dm(ks.b10)
        + (1 - F) / 3 * ns.qubits.ketutil.ket2dm(ks.b11)
    )
    return werner_state



# =============================================================================
# # -*- coding: utf-8 -*-
#
# Created on Tue Sep 10 14:06:26 2024
#
# @author: kenny
#
# =============================================================================
"""
An extensible library of `netsquid.nodes.connections.Connection <connections>`
 useful for a DQC and wrappers for creating instances of them.
"""

from netsquid.components import (
    ClassicalChannel,
    FibreDelayModel,
    QuantumChannel,
    QSource,
    SourceStatus,
)
from netsquid.nodes.connections import Connection, DirectConnection
from netsquid.qubits import StateSampler
from netsquid_physlayer.heralded_connection import MiddleHeraldedConnection


# DELETE COMMENTS BELOW ONCE FINISHED REFACTORING

# Changelog relative to original entangling connection (things likely to
# impact pre-existing code only):
#   1) removed keywork argument called `other_node_initial` (check this is)
#      not specified in any instances of the function
#   2) added state4distribution parameter
#   3) changed name of class itself

# Notes:
#   1) Be careful of syncronisation when using the
#      BlackBoxEntanglingQsourceConnection. This used to be handled via a
#      protocol on a central node containing a QSource but will now need to be
#      handled by ensuring that only one of the QPUs sends a trigger request
#      once they have both ensured they are ready to receive an entangled qubit
#      via appropriate handshake protocols

# TO DO:
#   Think about how to model delay. You may wish to allow user to specify
#   noise parameters to allow them to use the connection as abstract or
#   real as they wish


def create_classical_fibre_link(network, node_a, node_b, length, label="classical"):
    """
    Setus up a classical fibre connection between two QPUs.

    Parameters
    ----------
    network : :class:`~netsquid.nodes.network.Network`
        The entire network.
    node_a, node_b : :class:`~netsquid.nodes.Node`
        The QPUs to be linked.
    length : float
        the length of the connection.
    """
    connection = DirectConnection(
        "DirectConnection",
        channel_AtoB=ClassicalChannel(
            "Channel_A2B", length=length, models={"delay_model": FibreDelayModel()}
        ),
        channel_BtoA=ClassicalChannel(
            "Channel_B2A", length=length, models={"delay_model": FibreDelayModel()}
        ),
    )
    # generating names obeying netsquid naming conventions for port on QPUs that
    # connect a Node to a connection but using the (unique) node name as the ID
    node_a_port_name = node_a.connection_port_name(node_b.name, label=label)
    node_b_port_name = node_b.connection_port_name(node_a.name, label=label)
    network.add_connection(
        node_a,
        node_b,
        connection=connection,
        port_name_node1=node_a_port_name,
        port_name_node2=node_b_port_name,
        label=label,
    )


class BlackBoxEntanglingQsourceConnection(Connection):
    """
    Intended to connnect source of entangled photons and 2 QPUs.

    This class only works with the density matrix formalism and so violates the
    formalism agnosticism of NetSqiud. It generates a pair of entangled qubits
    sharing a particular state without
    regard for how that state was generated. The state is intended to be
    specified analytically and the state definition can be used to specify the
    noise on the entangled photons if desired. In this way, the need for
    multiple simulation runs (which would be required by a probabilistic noise
    model) is circumvented. Therefore distribution with this connection is
    deterministic and any probablistic effects are incorporated into the
    chosen entanglement distribution rate (ie, the average entanglement
    distribution rate should be used).

    The black box :class:`~netsquid.components.qsource.QSource` is triggered
    via the sending of a classical trigger message from either QPU. It is
    assumed that both QPUs are ready to receive an entangled qubit when the
    trigger message is sent, which will need to be enforced using the
    :class:`~netsquid.protocols.nodeprotocols.NodeProtocol` (s) used to send
    the trigger message. Only one trigger message should be sent per
    entanglement distribution desired to avoid unintended behaviour (ie, a
    trigger message should not be sent by both QPUs requesting the same
    entangled pair).

    Parameters
    ----------
    delay : float
        The time for entanglement distribution to occur
        [ns].
    state4distribution : array
        The density matrix of the state that the entangled qubits should
        possess.
    name : str, optional
        Name of this connection. Default is
        "BlackBoxEntanglingQsourceConnection".
    """

    def __init__(
        self, delay, state4distribution, name="BlackBoxEntanglingQsourceConnection"
    ):
        super().__init__(name=name)
        qchannel_qsource2A = QuantumChannel("qchannel_qsource2A", delay=delay)
        qchannel_qsource2B = QuantumChannel("qchannel_qsource2B", delay=delay)
        cchannelA2qsource_trigger = ClassicalChannel(
            "cchannel_A2qsource_trigger", delay=0
        )
        # the classical delay is
        # implicitly included in
        # delay of quantum
        # channels here to model
        # the total entanglment
        # distribution time.
        # TO DO: revisit this
        cchannelB2qsource_trigger = ClassicalChannel(
            "cchannel_B2qsource_trigger", delay=0
        )
        # TO DO: revisit delay
        # as commented above
        qsource = QSource(
            "qsource",
            StateSampler([state4distribution], [1.0]),
            num_ports=2,
            status=SourceStatus.EXTERNAL,
        )
        # adding extra trigger port and then forwareding that to the trigger so
        # that two classical connections can be connected to the trigger.
        qsource.add_ports(["extra_trigger"])
        qsource.ports["extra_trigger"].forward_input(qsource.ports["trigger"])
        # TO DO: tinker with qsource to see if it can be made to emit
        # multiple pairs of entangled photons
        # Adding channels and forwarding to and from the connection's
        # "A" and "B" ports to the input ("send") and ouput ("recv")
        # ports of the channel. An input to the channel's "send" port then
        # forwards to the
        # "recv" port in the channel behind the scenes, after some delay
        self.add_subcomponent(qchannel_qsource2A, forward_output=[("A", "recv")])
        self.add_subcomponent(qchannel_qsource2B, forward_output=[("B", "recv")])
        self.add_subcomponent(cchannelA2qsource_trigger, forward_input=[("A", "send")])
        self.add_subcomponent(cchannelB2qsource_trigger, forward_input=[("B", "send")])
        self.add_subcomponent(qsource)
        # adding the remaining connections between components within the
        # overall Connection
        qsource.ports["qout0"].connect(qchannel_qsource2A.ports["send"])
        qsource.ports["qout1"].connect(qchannel_qsource2B.ports["send"])
        qsource.ports["trigger"].connect(cchannelA2qsource_trigger.ports["recv"])
        qsource.ports["extra_trigger"].connect(cchannelB2qsource_trigger.ports["recv"])


class ProbabilisticQSourceConnection(Connection):
    """
    Intended to connnect source of entangled photons and 2 QPUs.

    Generates a pair of entangled qubits sharing a particular state without
    regard for how that state was generated.

    The black box :class:`~netsquid.components.qsource.QSource` is triggered
    via the sending of a classical trigger message from either QPU. It is
    assumed that both QPUs are ready to receive an entangled qubit when the
    trigger message is sent, which will need to be enforced using the
    :class:`~netsquid.protocols.nodeprotocols.NodeProtocol`(s) used to send
    the trigger message. Only one trigger message should be sent per
    entanglement distribution desired to avoid unintended behaviour (ie, a
    trigger message should not be sent by both QPUs requesting the same
    entangled pair).

    Parameters
    ----------
    delay : float
        The time for entanglement distribution to occur
        [ns].
    states4distribution : list of tuples (state, float)
        Where state is `~netsquid.qubits.qrepr.QRepr` or `numpy.ndarray`.
        Quantum state representations to distribute. Quantum state
        representations and the
        probability of being in that state
    name : str, optional
        Name of this connection. Default is
        "BlackBoxEntanglingQsourceConnection".
    """

    def __init__(
        self, delay, states4distribution, name="ProbabilisticQsourceConnection"
    ):
        super().__init__(name=name)
        qchannel_qsource2A = QuantumChannel("qchannel_qsource2A", delay=delay)
        qchannel_qsource2B = QuantumChannel("qchannel_qsource2B", delay=delay)
        cchannelA2qsource_trigger = ClassicalChannel(
            "cchannel_A2qsource_trigger", delay=0
        )
        # the classical delay is
        # implicitly included in
        # delay of quantum
        # channels here to model
        # the total entanglment
        # distribution time.
        # TO DO: revisit this
        cchannelB2qsource_trigger = ClassicalChannel(
            "cchannel_B2qsource_trigger", delay=0
        )
        # TO DO: revisit delay
        # as commented above
        qsource = QSource(
            "qsource",
            StateSampler(
                [state[0] for state in states4distribution],
                [state[1] for state in states4distribution],
            ),
            num_ports=2,
            status=SourceStatus.EXTERNAL,
        )
        # adding extra trigger port and then forwareding that to the trigger so
        # that two classical connections can be connected to the trigger.
        qsource.add_ports(["extra_trigger"])
        qsource.ports["extra_trigger"].forward_input(qsource.ports["trigger"])
        # TO DO: tinker with qsource to see if it can be made to emit
        # multiple pairs of entangled photons
        # Adding channels and forwarding to and from the connection's
        # "A" and "B" ports to the input ("send") and ouput ("recv")
        # ports of the channel. An input to the channel's "send" port then
        # forwards to the
        # "recv" port in the channel behind the scenes, after some delay
        self.add_subcomponent(qchannel_qsource2A, forward_output=[("A", "recv")])
        self.add_subcomponent(qchannel_qsource2B, forward_output=[("B", "recv")])
        self.add_subcomponent(cchannelA2qsource_trigger, forward_input=[("A", "send")])
        self.add_subcomponent(cchannelB2qsource_trigger, forward_input=[("B", "send")])
        self.add_subcomponent(qsource)
        # adding the remaining connections between components within the
        # overall Connection
        qsource.ports["qout0"].connect(qchannel_qsource2A.ports["send"])
        qsource.ports["qout1"].connect(qchannel_qsource2B.ports["send"])
        qsource.ports["trigger"].connect(cchannelA2qsource_trigger.ports["recv"])
        qsource.ports["extra_trigger"].connect(cchannelB2qsource_trigger.ports["recv"])


def create_black_box_central_source_entangling_link(
    network, node_a, node_b, state4distribution, ent_dist_rate=0
):
    """
    Sets up an abstract entangling link between QPUs.

    Adds BlackBoxEntanglingQsourceConnection between two QPUs.

    Parameters
    ----------
    network : :class:`~netsquid.nodes.network.Network`
        The entire network.
    node_a, node_b : :class:`~netsquid.nodes.Node`
        A network node.
    state4distribution : numpy.ndarray
        The entangled state distributed between nodes when
        an EPR pair is requested. Default is |phi^+> Bell state (formalism not
        fixed to ket)
    ent_dist_rate : float, optional
        The rate of entanglement distribution [Hz]. The default is 0
    Notes
    -----
    This abstracts from the details of photon generation by treating flying
    and communication qubits as the same thing. Restraints on the number of
    communication qubits can be enforced at the QPU nodes but entangled
    communication qubits are generated at a central quantum source and sent
    to the QPUs. In this way, we can model error and loss but needn't simulate
    the details of entanglement between static communication qubits and photons.
    """
    if ent_dist_rate > 0:
        ent_dist_time = 1e9 / ent_dist_rate
    elif ent_dist_rate == 0.0:
        ent_dist_time = 0
    else:
        raise ValueError(
            f"{ent_dist_rate} is not a valid entanglement "
            f"distribution rate. The rate must be >= 0"
        )
    # commented out block below is for after refactor
    connection = BlackBoxEntanglingQsourceConnection(
        delay=ent_dist_time, state4distribution=state4distribution
    )
    # generating names obeying netsquid naming conventions for port on QPUs that
    # connect a Node to a connection but using the (unique) node name as the ID
    node_a_port_name = node_a.connection_port_name(node_b.name, label="entangling")
    node_b_port_name = node_b.connection_port_name(node_a.name, label="entangling")
    network.add_connection(
        node_a,
        node_b,
        connection=connection,
        port_name_node1=node_a_port_name,
        port_name_node2=node_b_port_name,
        label="entangling",
    )


def create_bb_elink(network, node_a, node_b, state4distribution, ent_dist_rate=0):
    """
    Alias for create_black_box_central_source_entangling_link.

    Parameters
    ----------
    network : TYPE
        DESCRIPTION.
    node_a : TYPE
        DESCRIPTION.
    node_b : TYPE
        DESCRIPTION.
    state4distribution : TYPE
        DESCRIPTION.
    ent_dist_rate : TYPE, optional
        DESCRIPTION. The default is 0.

    Returns
    -------
    None.

    """
    create_black_box_central_source_entangling_link(
        network, node_a, node_b, state4distribution, ent_dist_rate=ent_dist_rate
    )


def create_probabilistic_qsource_connection(
    network, node_a, node_b, states4distribution, ent_dist_rate=0
):
    """
    Sets up an abstract entangling link between QPUs.

    Adds BlackBoxEntanglingQsourceConnection between two QPUs.

    Parameters
    ----------
    network : :class:`~netsquid.nodes.network.Network`
        The entire network.
    node_a, node_b : :class:`~netsquid.nodes.Node`
        A network node.
    states4distribution : dict
        Keys are quantum state representations `~netsquid.qubits.qrepr.QRepr`,
        values are probability of being in that state.
    ent_dist_rate : float
        The rate of entanglement distribution [Hz].
    Notes
    -----
    This abstracts from the details of photon generation by treating flying
    and communication qubits as the same thing. Restraints on the number of
    communication qubits can be enforced at the QPU nodes but entangled
    communication qubits are generated at a central quantum source and sent
    to the QPUs. In this way, we can model error and loss but needn't simulate
    the details of entanglement between static communication qubits and photons.
    """
    if ent_dist_rate > 0:
        ent_dist_time = 1e9 / ent_dist_rate
    elif ent_dist_rate == 0.0:
        ent_dist_time = 0
    else:
        raise ValueError(
            f"{ent_dist_rate} is not a valid entanglement "
            f"distribution rate. The rate must be >= 0"
        )
    # commented out block below is for after refactor
    connection = ProbabilisticQSourceConnection(
        delay=ent_dist_time, states4distribution=states4distribution
    )
    # generating names obeying netsquid naming conventions for port on QPUs that
    # connect a Node to a connection but using the (unique) node name as the ID
    node_a_port_name = node_a.connection_port_name(node_b.name, label="entangling")
    node_b_port_name = node_b.connection_port_name(node_a.name, label="entangling")
    network.add_connection(
        node_a,
        node_b,
        connection=connection,
        port_name_node1=node_a_port_name,
        port_name_node2=node_b_port_name,
        label="entangling",
    )


def create_midpoint_heralded_entangling_link(
    network,
    node_a,
    node_b,
    length=2e-3,
    p_loss_init=0,
    p_loss_length=0,
    speed_of_light=200000,
    dark_count_probability=0,
    detector_efficiency=1.0,
    visibility=1.0,
    num_resolving=False,
    coin_prob_ph_ph=1.0,
    coin_prob_ph_dc=1.0,
    coin_prob_dc_dc=1.0,
):
    """
    Sets up an entangling connection with a central BSM between two QPUs.

    Adds a
    :class:`netsquid_physlayer.heralded_connection.MiddleHeraldedConnection`
    between two nodes.

    Parameters
    ----------
    network : :class:`~netsquid.nodes.network.Network`
        The entire network.
    node_a, node_b : :class:`~netsquid.nodes.Node`
        The QPU nodes to be linked.
    length: float
        Total length [km] of heralded connection (i.e. sum of fibers on both sides on midpoint station).
    p_loss_init: float (optional)
        Probability that photons are lost when entering connection the connection on either side.
    speed_of_light: float (optional)
        Speed of light [km/s] in fiber on either side.
    p_loss_length: float (optional)
        Attenuation coefficient [dB/km] of fiber on either side.
    dark_count_probability: float (optional)
        Dark-count probability per detection
    detector_efficiency: float (optional)
        Probability that the presence of a photon leads to a detection event
    visibility: float (optional)
        Hong-Ou-Mandel visibility of photons that are being interfered (measure of photon indistinguishability)
    num_resolving : bool (optional)
        determines whether photon-number-resolving detectors are used for the Bell-state measurement
    coin_prob_ph_ph : float (optional)
        Coincidence probability for two photons. When using a coincidence time window in the double-click protocol,
        two clicks are only accepted if they occurred within one coincidence time window away from each other.
        This parameter is the probability that if both clicks are photon detections,
        they are within one coincidence window. In general, this depends not only on the size of the coincidence
        time window, but also on the state of emitted photons and the total detection time window. Defaults to 1.
    coin_prob_ph_dc : float (optional)
        Coincidence probability for a photon and a dark count.
        When using a coincidence time window in the double-click protocol,
        two clicks are only accepted if they occurred within one coincidence time window away from each other.
        This parameter is the probability that if one click is a photon detection and the other a dark count,
        they are within one coincidence window. In general, this depends not only on the size of the coincidence
        time window, but also on the state of emitted photons and the total detection time window. Defaults to 1.
    coin_prob_dc_dc : float (optional)
        Coincidence probability for two dark counts. When using a coincidence time window in the double-click protocol,
        two clicks are only accepted if they occurred within one coincidence time window away from each other.
        This parameter is the probability that if both clicks are dark counts,
        they are within one coincidence window. In general, this depends on the size of the coincidence time window
        and the total detection time window. Defaults to 1.

    References
    ----------
    Most of this docstring is taken (with some slight modification in
    places) from the docstring for
    `netsquid_physlayer.heralded_connection.MiddleHeraldedConnection`
    in accordance with the apache 2.0 license available at
    http://www.apache.org/licenses/LICENSE-2.0
    """
    connection = MiddleHeraldedConnection(
        name="MiddleHeraldedConnection",
        length=length,
        p_loss_init=p_loss_init,
        p_loss_length=p_loss_length,
        speed_of_light=speed_of_light,
        dark_count_probability=dark_count_probability,
        detector_efficiency=detector_efficiency,
        visibility=visibility,
        num_resolving=num_resolving,
        coin_prob_ph_ph=coin_prob_ph_ph,
        coin_prob_ph_dc=coin_prob_ph_dc,
        coin_prob_dc_dc=coin_prob_dc_dc,
    )
    # generating names obeying netsquid naming conventions for port on QPUs that
    # connect a Node to a connection but using the (unique) node name as the ID
    node_a_port_name = node_a.connection_port_name(node_b.name, label="entangling")
    node_b_port_name = node_b.connection_port_name(node_a.name, label="entangling")
    network.add_connection(
        node_a,
        node_b,
        connection=connection,
        port_name_node1=node_a_port_name,
        port_name_node2=node_b_port_name,
        label="entangling",
    )
    node_a.qmemory.ports["qout"].forward_output(node_a.ports[node_a_port_name])
    node_b.qmemory.ports["qout"].forward_output(node_b.ports[node_b_port_name])



# =============================================================================
# # -*- coding: utf-8 -*-
#
# Created on Tue Oct 17 14:13:35 2023
#
# @author: kenny
# =============================================================================

"""
Defining a convenience class for encapsulating meta info on DQC circuits.
"""


class DqcCircuit:
    """
    Convenience class for encapsulating meta info on DQC circuits

    Parameters
        ----------
        qregs : dict of dicts
            The quantum registers and their associated info. Subdicts
            should have the keys 'size', and 'starting_index', and integer
            values for each.
        cregs : dict
            The classical registers and their associated sizes.
        native_gates : dict
            The gates native to the processor upon which the DQC circuit
            will be enacted.
        ops : list of lists
            The operations (such as gates, initialisations or measurements)
            in the quantum circuit written in a way dqc_simulator can
            understand.
        qreg2node_lookup : dict or None, optional
            A mapping from qreg names to node names. This can be used to
            specify the partitioning of the circuit manually by associating
            each qreg with an appropriate node name
        circuit_type : str or None, optional
            Meta info indicating what is left to do to the circuit
             Can be 'monolithic', 'unpartitioned' 'prepped4partitioning',
             or 'partitioned'. TO DO: add more options like 'optimised'.
    """

    def __init__(
        self, qregs, cregs, native_gates, ops, qreg2node_lookup=None, circuit_type=None
    ):
        self.qregs = qregs
        self.cregs = cregs
        self.native_gates = native_gates
        self.ops = ops
        self.qreg2node_lookup = qreg2node_lookup
        self.circuit_type = circuit_type
        self.qubit_count = 0
        self.scheme = None  # if this is a str, then it is the scheme all
        # two-qubit gates will be conducted using
        self.node_sizes = dict()  # when processed should be dict with entries of
        # form node_name : integer value
        self.gate_macros = dict()

    def replace_qreg_names(self, node_0_name="placeholder", node_1_name="placeholder"):
        """Replaces node names in each gate_spec in ops with a placeholder
        indicating that partitioning should be done

        Parameters
            ----------
            node_1_name : str, optional
                The word to use for the first node in each gate_spec or
                an appropriate_placeholder"""
        for gate_spec in self.ops:
            gate_spec[2] = node_0_name
            if len(gate_spec) >= 5:
                gate_spec[4] = node_1_name

    # =============================================================================
    #     def _replace_qreg_names_with_placeholder(self):
    #         self.replace_qreg_names(node_0_name='placeholder',
    #                                  node_1_name='placeholder')
    #         self.circuit_type = 'prepped4partitioning'
    #
    #     def _convert2monolithic(self):
    #         """Replaces node names in each gate_spec in ops with the name of
    #            monolithic processor"""
    #         self.replace_qreg_names(node_0_name='monolithic_processor',
    #                                  node_1_name='monolithic_processor')
    #         self.circuit_type = 'monolithic'
    #
    #     def _specify_partition_manually(self):
    #         for gate_spec in self.ops:
    #             gate_spec[2] = self.qreg2node_lookup[gate_spec[2]]
    #             if len(gate_spec) >= 5:
    #                 gate_spec[4] = self.qreg2node_lookup[gate_spec[4]]
    #         self.circuit_type = 'partitioned'
    #
    #
    #     def qregs2nodes(self, conversion_strategy):
    #         """
    #         Converts qreg names to node names for all gate_spec elements of
    #         self.ops.
    #
    #         Parameters
    #         ----------
    #         conversion_strategy : str or None, optional
    #             How to partition the circuit. This is used by the to ascertain how
    #             to convert the qreg names to node names. Can be 'monolithic',
    #             'manual', or 'auto'.
    #                 'monolithic' : all gate_spec elements in self.ops have
    #                                all node names set to "monolithic_processor"
    #                 'manual' : the node names in all gate_spec elements will be
    #                            specified by the qreg2node_lookup. This
    #                            corresponds to a user-specified partitioning of
    #                            the circuit
    #                 'auto' : all node names are replaced with 'placeholder',
    #                         awaiting automated partitioning. MAY BE SUPERFLUOUS
    #         """
    #         circuit_type_converters = {
    #                    'auto' : self._replace_qreg_names_with_placeholder, #MAY BE SUPERFLUOUS
    #                    'monolithic' : self._convert2monolithic,
    #                    'manual' : self._specify_partition_manually
    #                    }
    #         converter_subroutine = circuit_type_converters[conversion_strategy]
    #         converter_subroutine()
    # =============================================================================

    def add_scheme_to_2_qubit_gates(self, scheme):
        """
        Specifies the scheme to be used for all two qubit gates.

        Parameters
        ----------
        scheme : str
            The scheme to be used for all remote two-qubit gates
        """
        self.scheme = scheme
        for gate_spec in self.ops:
            if len(gate_spec) == 5:  # if two-qubit gate without scheme specified
                node0 = gate_spec[2]
                node1 = gate_spec[4]
                if node0 != node1:  # if remote:
                    gate_spec.append(scheme)

    def lock_in_gate_specs(self):
        """Renders all elements of self.ops immutable. This is useful for
        protecting against accidental alteration of elements in self.ops.
        """
        for gate_spec in self.ops:
            gate_spec = tuple(gate_spec)


# =============================================================================
#     def partition(self, partitioner):
#         self = partitioner(self)
#
# =============================================================================
